
import re
import os
import warnings
import libsbml
import networkx

from .c_graph import Cgraph
from .low_deficiency_approach import LowDeficiencyApproach
from .mass_conservation_approach import MassConservationApproach
from .semi_diffusive_approach import SemiDiffusiveApproach
from .advanced_deficiency_approach import AdvancedDeficiencyApproach


class CRNT:
    """
    Class for managing CRNT methods.
    """
    def __init__(self, path):
        """
        Initialization of CRNT class.

        Parameters
        -----------
        path: string
            String representation of the path to the XML file.
        Example
        ---------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/sbml_file.xml")
        """
        self.__path = path
        self.__sbml_model = None

        self.__validate_path()
        # self.__read_sbml_model()
        reader = libsbml.SBMLReader()
        sbml_doc = reader.readSBMLFromFile(self.__path)
        self.__sbml_model = sbml_doc.getModel()
        self.__cgraph = Cgraph(self.__sbml_model)  # init Cgraph object

    def __validate_path(self):
        if not os.path.isfile(self.__path):
            raise Exception("file is missing")

    def __read_sbml_model(self):
        reader = libsbml.SBMLReader()
        sbml_doc = reader.readSBMLFromFile(self.__path)
        # This is just to warn user. Not considered as a failure at this point.
        if sbml_doc.validateSBML() > 0:
            warnings.warn("Warning, SBML did not pass validation!\nProceeding anyway.")
        if sbml_doc.checkConsistency() > 0:
            warnings.warn("Warning, SBML did not pass checking of consistency!\nProceeding anyway.")
        self.__sbml_model = sbml_doc.getModel()
        pass

    # public methods

    def plot_c_graph(self):
        """
        Generates a matplotlib plot for the C-graph of the network using the networkx.draw function with circular and
        Kamada Kawai layout.

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/sbml_file.xml")
        >>> network.plot_c_graph()
        """
        self.__cgraph.plot()

    def plot_save_c_graph(self):
        """
        Saves the matplotlib plot for the C-graph of the network using the networkx.draw function with circular and
        Kamada Kawai layout to the file network_cgraph.png

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/sbml_file.xml")
        >>> network.plot_save_c_graph()
        """
        self.__cgraph.plot_save()

    def print_c_graph(self):
        """
        Prints the reactions and reaction labels for the network.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> network.print_c_graph()
            Reaction graph of the form
            reaction -- reaction label:
            s1+s2 -> s3  --  re1
            s3 -> s1+s2  --  re1r
            s3 -> s6+s2  --  re2
            s6+s7 -> s16  --  re3
            s16 -> s6+s7  --  re3r
            s16 -> s7+s1  --  re4
            s1+s6 -> s15  --  re6
            s15 -> s1+s6  --  re6r
            s15 -> 2*s6  --  re8
        """
        self.__cgraph.print()

    def get_network_graphml(self):
        """
        Writes the NetworkX Digraph to the file network.graphml. Note that this generation only includes the names of
        the nodes, edges, and edge reaction names, it does not include other list attributes of the nodes and edges.

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/sbml_file.xml")
        >>> network.get_network_graphml()
        """
        c_graph = self.__cgraph.get_graph()
        g = networkx.DiGraph()
        g.add_nodes_from(c_graph.nodes)
        g.add_edges_from(c_graph.edges)

        # add reaction names to graph
        edge_names = []
        for i in c_graph.edges:
            if (i[1], i[0]) in c_graph.edges:
                edge_names.append({"label": c_graph.edges[i]["label"], 'Bend': True})
            else:
                edge_names.append({"label": c_graph.edges[i]["label"], 'Bend': False})

        zip_obj = zip([i for i in g.edges], edge_names)
        edge_dict = dict(zip_obj)
        networkx.set_edge_attributes(g, edge_dict)

        #add if a complex is a boundary condition
        node_bc = [{'BC': c_graph.nodes[i]["species_bc"]} for i in c_graph.nodes]
        zip_obj = zip([i for i in g.nodes], node_bc)
        node_dict = dict(zip_obj)
        networkx.set_node_attributes(g, node_dict)

        networkx.write_graphml(g, "network.graphml")

    def basic_report(self):
        """
        Prints out basic CRNT properties of the network.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> network.basic_report()
            Number of species: 7
            Number of complexes: 9
            Number of reactions: 9
            Network deficiency: 2
        """

        report = """
        Number of species: %i
        Number of complexes: %i
        Number of reactions: %i 
        Network deficiency: %i
        """ % (
            len(self.__cgraph.get_species()),
            len(self.__cgraph.get_complexes()),
            len(self.__cgraph.get_reactions()),
            self.__cgraph.get_deficiency()
        )
        print("")
        print(re.sub(r"^\s+", "", report, flags=re.MULTILINE))

    def get_low_deficiency_approach(self):
        """
        Initializes and creates an object for the class LowDeficiencyApproach for the CRNT object constructed.

        See also
        ---------
        crnt4sbml.LowDeficiencyApproach

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/sbml_file.xml")
        >>> approach = network.get_low_deficiency_approach()
        """
        return LowDeficiencyApproach(self.__cgraph)

    def get_mass_conservation_approach(self):
        """
        Initializes and creates an object for the class MassConservationApproach for the CRNT object constructed.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        See also
        ---------
        crnt4sbml.MassConservationApproach

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_mass_conservation_approach()
            Creating Equilibrium Manifold ...
            Elapsed time for creating Equilibrium Manifold: 0.8010799999999998
            Solving for species' concentrations ...
            Elapsed time for finding species' concentrations: 1.1221319999999997
        """
        ldh = LowDeficiencyApproach(self.__cgraph)
        if ldh.does_satisfy_any_low_deficiency_theorem():
            message = """
            Network satisfies one of the low deficiency theorems.
            One should not run the optimization-based methods.
            """
            print(re.sub(r"^\s+", "", message, flags=re.MULTILINE))
            return None
        else:
            return MassConservationApproach(self.__cgraph)

    def get_semi_diffusive_approach(self):
        """
        Initializes and creates an object for the class SemiDiffusiveApproach for the CRNT object constructed.

        See also
        ---------
        crnt4sbml.SemiDiffusiveApproach

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/semi_diffusive_sbml_file.xml")
        >>> approach = network.get_semi_diffusive_approach()
        """
        ldh = LowDeficiencyApproach(self.__cgraph)
        if ldh.does_satisfy_any_low_deficiency_theorem():
            message = """
            Network satisfies one of the low deficiency theorems.
            One should not run the optimization-based methods.
            """
            print(re.sub(r"^\s+", "", message, flags=re.MULTILINE))
            return None
        else:
            return SemiDiffusiveApproach(self.__cgraph)

    def get_advanced_deficiency_approach(self):
        """
        Placeholder for Advanced Deficiency Approach. Future version of crnt4sbml will include the implementation of
        the Higher Deficiency Algorithm.
        """
        ldh = LowDeficiencyApproach(self.__cgraph)
        if ldh.does_satisfy_any_low_deficiency_theorem():
            message = """
            Network satisfies one of the low deficiency theorems.
            One should not  run the advanced deficiency method.
            """
            print(re.sub(r"^\s+", "", message, flags=re.MULTILINE))
            return None
        else:
            return AdvancedDeficiencyApproach(self.__cgraph)

    def get_c_graph(self):
        """
        Allows access to the class C-graph for the constructed CRNT object. Returns C-graph object for the provided
        CRNT object.

        See also
        ---------
        crnt4sbml.Cgraph

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/sbml_file.xml")
        >>> c_graph = network.get_c_graph()
        """
        return self.__cgraph

    @staticmethod
    def get_physiological_range(for_what=None):
        """Obtains physiological ranges.

        Parameters
        -----------
        for_what: string
            Accepted values: "concentration", "complex formation", "complex dissociation", or "catalysis"

        Returns
        --------
        concentration:
            5e-1,5e5 pM
        complex formation:
            1e-8,1e-4  pM^-1s^-1
        complex dissociation:
            1e-5,1e-3 s^-1
        catalysis:
            1e-3,1 s^-1

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/sbml_file.xml")
        >>> network.get_physiological_range("concentration")
        """
        valid = {"concentration", "complex_formation", "complex_dissociation", "monomolecular"}
        if for_what is None:
            warnings.warn("Please provide the argument what the range should be provided for.")
            return None
        if for_what not in valid:
            raise ValueError("for_what argument must be one of %r." % valid)
        if for_what == "concentration":
            return 5e-1, 5e5
        if for_what == "complex_formation":
            return 1e-8, 1e-4
        if for_what == "complex_dissociation":
            return 1e-5, 1e-3
        if for_what == "catalysis":
            return 1e-3, 1e0



import re


class LowDeficiencyApproach:
    """
    Class for testing the Deficiency Zero and One Theorems.
    """
    def __init__(self, cgraph):
        """
        Initialization of LowDeficiency Approach class.

        See also
        ---------
        crnt4sbml.CRNT.get_low_deficiency_approach()
        """
        self.__satisfies_deficiency_zero_theorem = False
        self.__satisfies_deficiency_one_theorem = False
        self.__satisfies_relaxed_deficiency_one_theorem = False

        # key conditions
        self.__cond1 = cgraph.get_if_cgraph_weakly_reversible()
        cond2 = cgraph.get_deficiency() == 0
        lcd = cgraph.get_linkage_classes_deficiencies()
        cond3 = all([i < 2 for i in lcd])
        cond4 = sum(lcd) == cgraph.get_deficiency()
        tlc = cgraph.get_number_of_terminal_strong_lc_per_lc()
        cond5 = all([i < 2 for i in tlc])

        # decision tree for low deficiency theorems
        if cond2:
            # deficiency zero theorem
            self.__satisfies_deficiency_zero_theorem = True
        elif self.__cond1 and cond3 and cond4:
            # deficiency one theorem
            self.__satisfies_deficiency_one_theorem = True
        elif cond5 and cond3 and cond4:
            # deficiency one theorem with relaxed criteria
            self.__satisfies_relaxed_deficiency_one_theorem = True

    def does_satisfy_any_low_deficiency_theorem(self):
        """
        Function to see if the network satisfies the Deficiency Zero or One Theorem. Returns True if the network
        satisfies the Deficiency Zero or One Theorem, False otherwise.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_low_deficiency_approach()
        >>> print(approach.does_satisfy_any_low_deficiency_theorem())
            False
        """
        return (self.__satisfies_deficiency_zero_theorem or
                self.__satisfies_deficiency_one_theorem or
                self.__satisfies_relaxed_deficiency_one_theorem)

    def does_satisfy_deficiency_zero_theorem(self):
        """
        Function to see if the network satisfies the Deficiency Zero Theorem. Returns True if the network
        satisfies the Deficiency Zero Theorem, False otherwise.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_low_deficiency_approach()
        >>> print(approach.does_satisfy_deficiency_zero_theorem())
            False
        """
        return self.__satisfies_deficiency_zero_theorem

    def does_satisfy_deficiency_one_theorem(self):
        """
        Function to see if the network satisfies the Deficiency One Theorem. Returns True if the network
        satisfies the Deficiency One Theorem, False otherwise.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_low_deficiency_approach()
        >>> print(approach.does_satisfy_deficiency_one_theorem())
            False
        """
        return (self.__satisfies_deficiency_one_theorem or
                self.__satisfies_relaxed_deficiency_one_theorem)

    def report_deficiency_zero_theorem(self):
        """
        Prints out the applicability of the Deficiency Zero Theorem for the provided network.
        Possible output:

        "By the Deficiency Zero Theorem, the differential equations
        cannot admit a positive equilibrium or a cyclic composition
        trajectory containing a positive composition. Thus, multiple
        equilibria cannot exist for the network."

        or

        "By the Deficiency Zero Theorem, there exists within each positive
        stoichiometric compatibility class precisely one equilibrium.
        Thus, multiple equilibria cannot exist for the network."

        or

        "The network does not satisfy Deficiency Zero Theorem."

        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_low_deficiency_approach()
        >>> print(approach.report_deficiency_zero_theorem())
            The network does not satisfy Deficiency Zero Theorem.
        """
        if self.__satisfies_deficiency_zero_theorem and not self.__cond1:
            report = ("""
            By the Deficiency Zero Theorem, the differential equations 
            cannot admit a positive equilibrium or a cyclic composition 
            trajectory containing a positive composition. Thus, multiple 
            equilibria cannot exist for the network. 
            """)
            print(re.sub(r"^\s+", "", report, flags=re.MULTILINE))
        elif self.__satisfies_deficiency_zero_theorem and self.__cond1:
            report = (""" 
            By the Deficiency Zero Theorem, there exists within each positive
            stoichiometric compatibility class precisely one equilibrium.
            Thus, multiple equilibria cannot exist for the network.
            """)
            print(re.sub(r"^\s+", "", report, flags=re.MULTILINE))
        else:
            print("The network does not satisfy Deficiency Zero Theorem.")

    def report_deficiency_one_theorem(self):
        """
        Prints out the applicability of the Deficiency One Theorem for the provided network.
        Possible output:

        "By the Deficiency One Theorem, the differential equations
        admit precisely one equilibrium in each positive stoichiometric
        compatibility class. Thus, multiple equilibria cannot exist
        for the network."

        or

        "The network satisfies relaxed Deficiency One Theorem. That is it
        is not weakly reversable, but each linkage class contains no more
        than one terminal linkage class. There can exist within a positive
        stoichiometric compatibility class at most one equilibrium.
        Thus, multiple equilibria cannot exist for the network."

        or

        "The network does not satisfy Deficiency One Theorem."

        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_low_deficiency_approach()
        >>> print(approach.report_deficiency_zero_theorem())
            The network does not satisfy Deficiency One Theorem.
        """
        if self.__satisfies_deficiency_one_theorem:
            report = ("""
            By the Deficiency One Theorem, the differential equations
            admit precisely one equilibrium in each positive stoichiometric
            compatibility class. Thus, multiple equilibria cannot exist 
            for the network.
            """)
            print(re.sub(r"^\s+", "", report, flags=re.MULTILINE))
        elif self.__satisfies_relaxed_deficiency_one_theorem:
            report = ("""
            The network satisfies relaxed Deficiency One Theorem. That is it 
            is not weakly reversable, but each linkage class contains no more 
            than one terminal linkage class. There can exist within a positive 
            stoichiometric compatibility class at most one equilibrium.
            Thus, multiple equilibria cannot exist for the network.
            """)
            print(re.sub(r"^\s+", "", report, flags=re.MULTILINE))
        else:
            print("The network does not satisfy Deficiency One Theorem.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `crnt4sbml` package."""

import unittest
import crnt4sbml
import pickle
import io
import contextlib


class TestCRNT4SBML_Test(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        mass_con_crnt_obj = crnt4sbml.CRNT("./sbml_files/Fig1Ci.xml")
        cls.mass_con_c_graph_list = cls.generate_c_graph_list(mass_con_crnt_obj, True)
        cls.mass_con_low_def_list = cls.generate_low_def_list(mass_con_crnt_obj)
        cls.mass_con_list = cls.generate_mass_con_list(mass_con_crnt_obj)

        semi_diff_crnt_obj = crnt4sbml.CRNT("./sbml_files/Fig1Cii.xml")
        cls.semi_diff_c_graph_list = cls.generate_c_graph_list(semi_diff_crnt_obj, False)
        cls.semi_diff_low_def_list = cls.generate_low_def_list(semi_diff_crnt_obj)
        cls.semi_diff_list = cls.generate_semi_diff_list(semi_diff_crnt_obj)

        low_def_crnt_obj = crnt4sbml.CRNT("./sbml_files/feinberg_ex3_13.xml")
        cls.low_def_c_graph_list = cls.generate_c_graph_list(low_def_crnt_obj, False)
        cls.low_def_low_def_list = cls.generate_low_def_list(low_def_crnt_obj)

    @staticmethod
    def generate_c_graph_list(crnt_obj, flag):
        a = crnt_obj.get_c_graph().get_a()
        complexes = crnt_obj.get_c_graph().get_complexes()
        deficiency = crnt_obj.get_c_graph().get_deficiency()
        dim_eq_man = crnt_obj.get_c_graph().get_dim_equilibrium_manifold()
        g_edges = crnt_obj.get_c_graph().get_g_edges()
        g_nodes = crnt_obj.get_c_graph().get_g_nodes()
        reversible = crnt_obj.get_c_graph().get_if_cgraph_weakly_reversible()
        link_def = crnt_obj.get_c_graph().get_linkage_classes_deficiencies()
        classification = crnt_obj.get_c_graph().get_network_dimensionality_classification()
        num_term = crnt_obj.get_c_graph().get_number_of_terminal_strong_lc_per_lc()
        ode_system = crnt_obj.get_c_graph().get_ode_system()
        psi = crnt_obj.get_c_graph().get_psi()
        reactions = crnt_obj.get_c_graph().get_reactions()
        s = crnt_obj.get_c_graph().get_s()
        species = crnt_obj.get_c_graph().get_species()
        revers_link = crnt_obj.get_c_graph().get_weak_reversibility_of_linkage_classes()
        y = crnt_obj.get_c_graph().get_y()

        if flag:
            b = crnt_obj.get_c_graph().get_b()
            lambda_val = crnt_obj.get_c_graph().get_lambda()

            return [a, b, complexes, deficiency, dim_eq_man, g_edges, g_nodes, reversible, link_def, classification,
                    num_term, ode_system, psi, reactions, s, species, revers_link, y, lambda_val]
        else:
            return [a, complexes, deficiency, dim_eq_man, g_edges, g_nodes, reversible, link_def, classification,
                    num_term, ode_system, psi, reactions, s, species, revers_link, y]

    @staticmethod
    def generate_low_def_list(crnt_obj):

        d_s_any = crnt_obj.get_low_deficiency_approach().does_satisfy_any_low_deficiency_theorem()
        d_s_one = crnt_obj.get_low_deficiency_approach().does_satisfy_deficiency_one_theorem()
        d_s_zero = crnt_obj.get_low_deficiency_approach().does_satisfy_deficiency_zero_theorem()

        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            crnt_obj.get_low_deficiency_approach().report_deficiency_one_theorem()
        report_one = f.getvalue()
        f.close()

        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            crnt_obj.get_low_deficiency_approach().report_deficiency_zero_theorem()
        report_zero = f.getvalue()
        f.close()

        return [d_s_any, d_s_one, d_s_zero, report_one, report_zero]

    @staticmethod
    def generate_mass_con_list(crnt_obj):

        approach = crnt_obj.get_mass_conservation_approach()
        concentration_solutions = approach.get_concentration_solutions()
        concentration_vals = approach.get_concentration_vals()
        conservation_laws = approach.get_conservation_laws()
        dch_matrix = approach.get_dch_matrix()
        decision_vector = approach.get_decision_vector()
        g_matrix = approach.get_g_matrix()
        h_vector = approach.get_h_vector()
        objective_fun_params = approach.get_objective_fun_params()
        symbolic_objective_fun = approach.get_symbolic_objective_fun()
        w_matrix = approach.get_w_matrix()
        w_nullspace = approach.get_w_nullspace()

        return [concentration_solutions, concentration_vals, conservation_laws, dch_matrix, decision_vector, g_matrix,
                h_vector, objective_fun_params, symbolic_objective_fun, w_matrix, w_nullspace]

    @staticmethod
    def generate_semi_diff_list(crnt_obj):

        approach = crnt_obj.get_semi_diffusive_approach()
        boundary_species = approach.get_boundary_species()
        decision_vector = approach.get_decision_vector()
        key_species = approach.get_key_species()
        mu_vector = approach.get_mu_vector()
        non_key_species = approach.get_non_key_species()
        s_to_matrix = approach.get_s_to_matrix()
        symbolic_objective_fun = approach.get_symbolic_objective_fun()
        symbolic_polynomial_fun = approach.get_symbolic_polynomial_fun()
        y_r_matrix = approach.get_y_r_matrix()

        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            approach.print_decision_vector()
        print_decision_vector = f.getvalue()
        f.close()

        return [boundary_species, decision_vector, key_species, non_key_species, mu_vector, s_to_matrix,
                symbolic_objective_fun, symbolic_polynomial_fun, y_r_matrix, print_decision_vector]

    def test_mass_conservation_c_graph(self):

        with open('./tests/mass_con_c_graph_list.pickle', 'rb') as pick_file:
            mass_con_c_graph_list = pickle.loads(pick_file.read())

        for i in range(19):
            assert(self.mass_con_c_graph_list[i] == mass_con_c_graph_list[i])

    def test_semi_diffusive_c_graph(self):

        with open('./tests/semi_diff_c_graph_list.pickle', 'rb') as pick_file:
            semi_diff_c_graph_list = pickle.loads(pick_file.read())

        for i in range(17):
            assert(self.semi_diff_c_graph_list[i] == semi_diff_c_graph_list[i])

    def test_low_deficiency_c_graph(self):

        with open('./tests/low_def_c_graph_list.pickle', 'rb') as pick_file:
            low_def_c_graph_list = pickle.loads(pick_file.read())

        for i in range(17):
            assert(self.low_def_c_graph_list[i] == low_def_c_graph_list[i])

    def test_mass_conservation_low_def(self):

        with open('./tests/mass_con_low_def_list.pickle', 'rb') as pick_file:
            mass_con_low_def_list = pickle.loads(pick_file.read())

        for i in range(5):
            assert(self.mass_con_low_def_list[i] == mass_con_low_def_list[i])

    def test_semi_diffusive_low_def(self):

        with open('./tests/semi_diff_low_def_list.pickle', 'rb') as pick_file:
            semi_diff_low_def_list = pickle.loads(pick_file.read())

        for i in range(5):
            assert(self.semi_diff_low_def_list[i] == semi_diff_low_def_list[i])

    def test_low_deficiency_low_def(self):

        with open('./tests/low_def_low_def_list.pickle', 'rb') as pick_file:
            low_def_low_def_list = pickle.loads(pick_file.read())

        for i in range(5):
            assert(self.low_def_low_def_list[i] == low_def_low_def_list[i])

    def test_mass_conservation(self):

        with open('./tests/mass_con_list.pickle', 'rb') as pick_file:
            mass_con_list = pickle.loads(pick_file.read())

        for i in range(11):
            assert(self.mass_con_list[i] == mass_con_list[i])

    def test_semi_diffusive(self):

        with open('./tests/semi_diff_list.pickle', 'rb') as pick_file:
            semi_diff_list = pickle.loads(pick_file.read())

        for i in range(10):
            assert(self.semi_diff_list[i] == semi_diff_list[i])


if __name__ == '__main__':
    unittest.main()




import sympy
import sys
import numpy
import networkx

mach_eps = numpy.finfo(float).eps


class AdvancedDeficiencyApproach:

    def __init__(self, cgraph):
        self.__cgraph = cgraph

        # construct dictionary for those edges that are reversible
        temp = [[e[0].split('+'), e[1].split('+')] for e in self.__cgraph.get_g_edges()]
        edges = [e for e in self.__cgraph.get_g_edges()]
        for i in temp:
            for j in i:
                j.sort()

        self.__reversible_dict = {}
        for i in range(len(edges)):

            if [temp[i][1], temp[i][0]] in temp:
                self.__reversible_dict.update({edges[i]: True})
            else:
                self.__reversible_dict.update({edges[i]: False})

        self.__Y = self.__cgraph.get_y()

        self.__species = self.__cgraph.get_species()

        self.__dict_id_name = self.__cgraph.get_dict_id_name()

    def __create_s_matrix_nullspace(self):
        S = sympy.zeros(len(self.__species), len(self.__orientation_graph_edges))
        original_nodes = self.__cgraph.get_g_nodes()
        for i in range(len(self.__orientation_graph_edges)):
            edge = self.__orientation_graph_edges[i]
            target = original_nodes.index(edge[1])
            source = original_nodes.index(edge[0])
            S[:, i] = self.__Y[:, target] - self.__Y[:, source]


        print("len(species)")
        print(len(self.__species))
        print("S")
        sympy.pprint(S)

        the_temp_null = S.nullspace()

        self.__d = len(the_temp_null)

        SN_row = len(self.__orientation_graph_edges)
        SN_col = len(the_temp_null)

        the_S_null = []
        for j in range(SN_row):
            the_S_null.append([the_temp_null[i][j] for i in range(SN_col)])

        return the_S_null, SN_row, SN_col

    def __initialize_orientation(self):

        c_graph = self.__cgraph.get_graph()

        pruned_edges = []
        for edge in self.__cgraph.get_g_edges():
            if (edge[1], edge[0]) not in pruned_edges:
                pruned_edges.append(edge)

        self.__orientation_graph = c_graph.edge_subgraph(pruned_edges).copy()
        self.__orientation_graph_edges = pruned_edges

    def __is_scalar_multiple_greater_than_zero(self, a, b):

        # get entries in a and b that have zero entry
        a_ind = [i for i in range(len(a)) if abs(a[i]) <= mach_eps]
        b_ind = [i for i in range(len(b)) if abs(b[i]) <= mach_eps]

        if a_ind == b_ind:
            vals = [b[i] / a[i] for i in range(len(a)) if i not in a_ind]
            if vals[0] > mach_eps and vals.count(vals[0]) == abs(len(a) - len(a_ind)):
                return True
            else:
                return False
        else:
            return False

    def __find_equivalence_classes(self):

        self.__the_S_null, SN_row, SN_col = self.__create_s_matrix_nullspace()

        self.__equivalence_classes = []
        self.__equivalence_classes_edges = []
        temp_list = [0.0 for i in range(SN_col)]

        # forming P_0 equivalence class
        temp_P_0 = []
        for i in range(SN_row):
            if all([abs(self.__the_S_null[i][j] - temp_list[j]) <= mach_eps for j in range(SN_col)]):
                temp_P_0.append(self.__orientation_graph_edges[i])

        if temp_P_0:
            self.__equivalence_classes.append(self.__orientation_graph.edge_subgraph(temp_P_0).copy())
            self.__equivalence_classes_edges.append(temp_P_0)
        else:
            self.__equivalence_classes.append("Empty")
            self.__equivalence_classes_edges.append("Empty")

        # getting dependent rows of Ker(L_orientation)
        # to create nonzero equivalent classes
        self.__dep_rows_S_null = []
        loop_iter = [i for i in range(SN_row) if i not in temp_P_0]
        for i in loop_iter:
            temp = [self.__orientation_graph_edges[i]]
            for j in loop_iter:
                if j != i:
                    temp_mat = sympy.Matrix([self.__the_S_null[i], self.__the_S_null[j]])
                    _, vals = temp_mat.T.rref()

                    if len(vals) == 1:
                        temp.append(self.__orientation_graph_edges[j])

                temp.sort()
                if temp not in self.__dep_rows_S_null and len(temp) != 1:
                    self.__dep_rows_S_null.append(temp)

        # creating nonzero equivalent classes
        [self.__equivalence_classes.append(self.__orientation_graph.edge_subgraph(self.__dep_rows_S_null[i]).copy())
         for i in range(len(self.__dep_rows_S_null))]

        [self.__equivalence_classes_edges.append(self.__dep_rows_S_null[i]) for i in range(len(self.__dep_rows_S_null))]

        #print("equivalence_classes_edges")
        #for i in self.__equivalence_classes_edges:
        #    print(i)

        #print("")

    def __equivalence_classes_check(self):

        # check that all reactions in P_0 are reversible
        if self.__equivalence_classes[0] == "Empty":
            a_chk = True
        else:
            is_reversible = []
            for i in self.__equivalence_classes_edges[0]:
                is_reversible.append(self.__reversible_dict[i])
            a_chk = all(is_reversible)

        b_chk = []
        for i in range(1, len(self.__equivalence_classes)):

            temp = []
            for j in self.__equivalence_classes_edges[i]:
                temp.append(self.__reversible_dict[j])

            temp_f_ind = [ii for ii in range(len(temp)) if temp[ii] is False]

            if len(temp_f_ind) > 1:
                temp_scalar = []
                new_iter = [self.__orientation_graph_edges.index(self.__dep_rows_S_null[i - 1][k]) for k in
                            range(len(self.__dep_rows_S_null[i - 1])) if k in temp_f_ind]

                for ii in new_iter:
                    temp_scalar += [
                        self.__is_scalar_multiple_greater_than_zero(self.__the_S_null[ii], self.__the_S_null[jj]) for jj
                        in new_iter if jj != ii]

                b_chk.append(all(temp_scalar))

            else:
                b_chk.append(True)

        if all(b_chk) and a_chk:
            return True
        else:
            return False

    def __find_fundamental_classes(self):

        self.__fundamental_classes = []
        self.__fundamental_classes_edges = []
        for i in range(len(self.__equivalence_classes)):

            if self.__equivalence_classes[i] == "Empty":
                self.__fundamental_classes.append("Empty")
                self.__fundamental_classes_edges.append("Empty")
            else:
                self.__fundamental_classes.append(self.__equivalence_classes[i].copy())
                self.__fundamental_classes_edges.append(self.__equivalence_classes_edges[i][:])

                for j in range(len(self.__fundamental_classes_edges[i])):
                    if self.__reversible_dict[self.__fundamental_classes_edges[i][j]]:
                        self.__fundamental_classes[i].add_edge(self.__fundamental_classes_edges[i][j][1],
                                                               self.__fundamental_classes_edges[i][j][0])
                        self.__fundamental_classes_edges[i].append((self.__fundamental_classes_edges[i][j][1],
                                                                    self.__fundamental_classes_edges[i][j][0]))

    def __find_colinkage_sets(self):

        fund_len = len(self.__fundamental_classes)

        if self.__fundamental_classes[0] == "Empty":
            start = 1
        else:
            start = 0

        colinkage_sets = []
        strong_colinkage_sets = []
        for fund_i in range(start, fund_len):

            linkages = [self.__fundamental_classes[fund_i].subgraph(c) for c in
                        networkx.weakly_connected_components(self.__fundamental_classes[fund_i])]

            for lc_i in linkages:

                colinkage_sets.append([i for i in self.__cgraph.get_g_nodes() if i in lc_i.nodes])
                strong_lcs = [lc_i.subgraph(c).copy() for c in networkx.strongly_connected_components(lc_i)]

                temp2 = []
                for slc_j in strong_lcs:

                    slc_j_lbls = [i for i in self.__cgraph.get_g_nodes() if i in slc_j.nodes]

                    x = [lc_i.successors(v_i) for v_i in slc_j_lbls]
                    connected_out_vertices = [item for sublist in x for item in sublist]
                    d = set(connected_out_vertices) - set(slc_j_lbls)

                    if d == set():
                        temp2.append([slc_j_lbls, "T"])
                    else:
                        temp2.append([slc_j_lbls, "NT"])

                strong_colinkage_sets.append(temp2)

        return strong_colinkage_sets, colinkage_sets

    def __find_representative_set_w(self):

        self.__representative_set_W = []
        for P_i in self.__equivalence_classes_edges:
            if P_i != "Empty":
                P_i_reversibility = [self.__reversible_dict[i] for i in P_i]
                if all(P_i_reversibility):
                    self.__representative_set_W.append(P_i[0])
                else:
                    False_ind = P_i_reversibility.index(False)
                    self.__representative_set_W.append(P_i[False_ind])
            else:
                self.__representative_set_W.append("Empty")

    def __realign_orientation(self):

        #print("hi")

        #print(self.__orientation_graph_edges)

        #print("")
        realignment_flag = False
        for i in range(len(self.__representative_set_W)):
            if self.__representative_set_W[i] != "Empty":
                y_i = self.__representative_set_W[i]
                v_i = self.__orientation_graph_edges.index(y_i)
                for j in range(len(self.__equivalence_classes_edges[i])):
                    v_y = self.__orientation_graph_edges.index(self.__equivalence_classes_edges[i][j])
                    out = self.__is_scalar_multiple_greater_than_zero(self.__the_S_null[v_i], self.__the_S_null[v_y])
                    if not out:
                        print("attempting to realign orientation look into this!")
                        self.__orientation_graph.remove_edge(*self.__equivalence_classes_edges[i][j])
                        self.__orientation_graph_edges.remove(self.__equivalence_classes_edges[i][j])
                        self.__orientation_graph.add_edge(self.__equivalence_classes_edges[i][j][1],
                                                          self.__equivalence_classes_edges[i][j][0])
                        self.__orientation_graph_edges.append((self.__equivalence_classes_edges[i][j][1],
                                                               self.__equivalence_classes_edges[i][j][0]))

                        self.g_print(self.__cgraph.get_graph(), self.__orientation_graph_edges)

                        realignment_flag = True

                        sys.exit()

        if realignment_flag:
            self.__find_equivalence_classes()
            if not self.__equivalence_classes_check():
                "Network does not have the capacity for multiple steady states."
                sys.exit()

            # testing the realignment
            for i in range(len(self.__representative_set_W)):
                if self.__representative_set_W[i] != "Empty":
                    y_i = self.__representative_set_W[i]
                    v_i = self.__orientation_graph_edges.index(y_i)
                    for j in range(len(self.__equivalence_classes_edges[i])):
                        v_y = self.__orientation_graph_edges.index(self.__equivalence_classes_edges[i][j])
                        out = self.__is_scalar_multiple_greater_than_zero(self.__the_S_null[v_i], self.__the_S_null[v_y])
                        if not out:
                            print("After realignment criteria in step 5 not satisfied!")
                            sys.exit()

    def __create_basis_ker_L_O_intersect_Gam_W(self):

        print("self.__orientation_graph_edges")
        print(self.__orientation_graph_edges)

        temp_representative = self.__orientation_graph_edges

        indices = [i for i in range(len(self.__orientation_graph_edges)) if self.__orientation_graph_edges[i]
                   in self.__representative_set_W]

        print("indices")
        print(indices)

        basis_Gam_W = sympy.zeros(len(self.__orientation_graph_edges), len(indices))
        count = 0
        for i in indices:
            basis_Gam_W[i, count] = 1
            count += 1

        #print("basis_Gam_W")
        #sympy.pprint(basis_Gam_W)

        #sys.exit()

        #print("temp_representative")
        #print(temp_representative)
        #empty_indices = [i for i in range(len(temp_representative)) if temp_representative[i] == 'Empty']

        #remove empty entries
        #for i in reversed(empty_indices):
        #    del temp_representative[i]

        #del temp_representative[1]
        #print(temp_representative)

        S = sympy.zeros(len(self.__species), len(temp_representative))
        original_nodes = self.__cgraph.get_g_nodes()
        for i in range(len(temp_representative)):
            edge = temp_representative[i]
            target = original_nodes.index(edge[1])
            source = original_nodes.index(edge[0])
            S[:, i] = self.__Y[:, target] - self.__Y[:, source]


        #print("self.__the_S_null")
        #sympy.pprint(self.__the_S_null)


        #print("S")
        #sympy.pprint(S)

        #find a basis for R(L_O^T) i.e. column space of L_O^T
        S_T = S.T
        #print("S_T")
        #sympy.pprint(S_T)
        rref_matrix, rref_pivots = S_T.rref()

        #print("rref_pivots")
        #sympy.pprint(rref_pivots)

        basis_column_space_L_O_T = sympy.zeros(len(temp_representative), len(rref_pivots))
        count = 0
        for i in rref_pivots:
            #print(i)
            basis_column_space_L_O_T[:, count] = S_T[:, i]
            count += 1

        #print("basis_column_space_L_O_T")
        #sympy.pprint(basis_column_space_L_O_T)

        A = basis_column_space_L_O_T.row_join(-1*basis_Gam_W)

        #print("")

        #print("A")
        #sympy.pprint(A)

        #print("nullspace")
        A_null = A.nullspace()
        #print(A_null)

        if len(A_null) != ((len(self.__representative_set_W) - 1) - self.__d):
            print("q != w - d")
            sys.exit()

        #print("")
        #print(A_null[0])

        A_null_mat = sympy.zeros(len(rref_pivots), len(A_null))

        for i in range(len(A_null)):
            for j in range(len(rref_pivots)):
                A_null_mat[j, i] = A_null[i][j]

        #print("A_mat_null")
        #sympy.pprint(A_null_mat)

        #print("")
        #print("basis_column_space_L_O_T*A_null_mat")
        full_basis = basis_column_space_L_O_T*A_null_mat
        #sympy.pprint(full_basis)
        #print(full_basis.shape[1])
        full_basis_W = sympy.zeros(len(indices), full_basis.shape[1])
        #print("")
        count = 0
        for i in indices:
            full_basis_W[count, :] = full_basis[i, :]
            count += 1

        #temp_full_basis = full_basis_W[5,:].col_join(full_basis_W[0:5,:])

        #print("")
        #sympy.pprint(full_basis_W)
        #sympy.pprint(temp_full_basis)

        #sympy.pprint(sympy.Matrix([[0], [1], [-1], [0], [1], [-1]]))

        #temp = (temp_full_basis.row_join(sympy.Matrix([[1], [0], [0], [-1], [1], [0]]))).rref()

        #sympy.pprint(temp)


        return full_basis_W, indices


    def __forest_basis_exists(self, full_basis_W, indices):

        print("full_basis_W")
        sympy.pprint(full_basis_W)
        print("full basis W rows")
        basis_rows = [self.__orientation_graph_edges[i][0] + '->' + self.__orientation_graph_edges[i][1]
                      for i in indices]

        basis_columns = ['a' + str(i) for i in range(full_basis_W.shape[1])]

        print("basis_rows")
        print(basis_rows)
        print("basis_columns")
        print(basis_columns)
        edges = []
        for i in range(full_basis_W.shape[0]):
            for j in range(full_basis_W.shape[1]):
                if abs(full_basis_W[i,j]) > mach_eps:
                    edges.append((basis_rows[i], basis_columns[j]))

        basis_graph = networkx.DiGraph()

        print("edges")
        print(edges)

        basis_graph.add_edges_from(edges)

        print(networkx.is_forest(basis_graph))


    def change_name(self, edge):

        splits = edge.split('+')
        switched = [self.__dict_id_name[i] for i in splits]
        if len(switched) == 1:
            return switched[0]
        else:
            return switched[0] + '+' + switched[1]

    def g_print(self, graph, graph_edges):
        print("graph_edges")
        print(graph_edges)
        for e in graph_edges:
            if e == 'Empty':
                print(e)
            else:
                print(self.change_name(e[0]) + ' -> ' + self.change_name(e[1]) + '  --  ' + graph.edges[e]['label'])
        print("")

    def run_higher_deficiency_algorithm(self):

        #self.__cgraph.print()

        self.__initialize_orientation()
        self.g_print(self.__cgraph.get_graph(), self.__orientation_graph_edges)

        self.__find_equivalence_classes()

        #print("equivalence_classes")
        #for i in range(len(self.__equivalence_classes)):
        #    if i != 0:
        #        self.g_print(self.__equivalence_classes[i], self.__equivalence_classes_edges[i])
        #print("")

        if self.__equivalence_classes_check():
            self.__find_fundamental_classes()

            #print("")
            #print("fundamental classes")
            #for i in range(1, len(self.__fundamental_classes)):
            #    self.g_print(self.__cgraph.get_graph(), self.__fundamental_classes_edges[i])
            #print("")

            self.__strong_colinkage_sets, self.__colinkage_sets = self.__find_colinkage_sets()

            #print("strong_colinkage_sets")
            #for i in strong_colinkage_sets:
            #    for j in i:
            #        print(j)
            #    print("")
            #print("")

            #print("colinkage_sets")
            #for i in colinkage_sets:
            #    print(i)
            #print("")

            self.__find_representative_set_w()

            print("self.__representative_set_W")
            print(self.__representative_set_W)

            self.__realign_orientation()

            full_basis_W, indices = self.__create_basis_ker_L_O_intersect_Gam_W()

            self.__forest_basis_exists(full_basis_W, indices)

        else:
            "Network does not have the capacity for multiple steady states."
            sys.exit()

    def report(self):
        print("placeholder")



import os
import numpy
import sympy
import sympy.utilities.lambdify
import sys 
import time
import numpy.linalg
import itertools
import warnings 
from .bistability_finder import BistabilityFinder


# TODO: see why these imports produce
# segmentation fault: 11
##########################
# import pylab as plt
# import tellurium as te
# from rrplugins import
##########################

class MassConservationApproach:
    """
    Class for constructing variables and methods needed for the mass conservation approach.
    """
    def __init__(self, cgraph):
        """
        Initialization of the MassConservationApproach class.

        See also
        ---------
        crnt4sbml.CRNT.get_mass_conservation_approach()
        """
        self.__cgraph = cgraph

        if not all([i <= 1 for i in self.__cgraph.get_number_of_terminal_strong_lc_per_lc()]):
            print("The network is not uniterminal!")
            sys.exit()

        if not self.__cgraph.get_dim_equilibrium_manifold() > 0:
            print("# of species - rank(S) is not greater than zero!")
            print("The deficiency approach cannot be ran!")
            sys.exit()

        # declare key fields
        self.__deficiency_pars = None
        self.__concentration_pars = None
        self.__reaction_pars = None 
        self.__W = None  # matrix
        self.__W_nullspace = None
        self.__H = None  # vector
        self.__G = None  # matrix
        self.__symbolic_objective_fun = None
        self.__concentration_vals = None
        self.__decision_vector_x = None
        self.__concentration_funs = None 
        self.__objective_fun_params = None
        self.__lambda_objective_fun = None 
        self.__important_info = ""
        self.__numpy_dtype = None

        # vars used frequently
        self.__N = len(self.__cgraph.get_species())
        self.__R = len(self.__cgraph.get_reactions())
        self.__species = self.__cgraph.get_species()
        self.__reactions = self.__cgraph.get_reactions()
        self.__delta = self.__cgraph.get_deficiency()
        self.__M = len(self.__cgraph.get_complexes())
        self.__ell = len(self.__cgraph.get_linkage_classes())
        self.__lambda = self.__cgraph.get_dim_equilibrium_manifold()
        self.__classification = self.__cgraph.get_network_dimensionality_classification()
        # compute necessary vectors and matrices
        self.__create_deficiency_pars()
        self.__create_concentration_pars()
        self.__create_reaction_pars()
        self.__create_w_matrix()
        self.__create_w_nullspace()
        self.__create_h_vector()
        self.__create_g_matrix()
        self.__create_symbolic_objective_fun()
        self.__create_concentration_values()
        self.__create_decision_vector_x()
        self.__create_concentration_bounds_species()
        self.__create_concentration_lambda_fun()
        self.__create_objective_fun__lambda_fun()
        self.__create_g_matrix_lambda_fun()
        self.__create_dch_matrix_lambda_fun()

    def __create_deficiency_pars(self):
        # creating a vector of the deficiency parameters
        # \alpha_1, ..., \alpha_\delta
        self.__deficiency_pars = [sympy.symbols('a' + str(i + 1), real=True) for i
                                  in range(self.__delta)]

    def __create_concentration_pars(self):
        # putting the species in a list for more readable code
        self.__concentration_pars = [sympy.Symbol(self.__species[i], positive=True) for i in range(self.__N)]

    def __create_reaction_pars(self):
        self.__reaction_pars = [sympy.Symbol(self.__reactions[i], positive=True) for i in range(self.__R)]
        
    def __create_w_matrix(self):
        # concatenating Y and Lambda_T columnwise
        # to create [Y,Lambda_T]^T # TODO: is W transposed?
        self.__W = self.__cgraph.get_y().col_join(self.__cgraph.get_lambda().T)

    def __create_w_nullspace(self):
        # Finding the null space of [Y,Lambda_T]^T,
        # i.e. \omega_i i=1,...,\delta
        self.__W_nullspace = self.__W.nullspace()

    def __create_h_vector(self):

        print("Creating Equilibrium Manifold ...")
        start = time.process_time()
        # creates the H vector in equation (10) by finding the linearly
        # independent rows of (9)
        # symbolic for of psi vector
        psi_symbol_vec = sympy.zeros(self.__M, 1)
        for i in range(self.__M):
            psi_symbol_vec[i] = sympy.Symbol('psi' + str(i))

        # creating the right-hand side defined by
        # \sum_{i=1}^\delta \alpha_i \omega_i
        rhs = sympy.zeros(self.__M, 1)
        temp = sympy.zeros(self.__M, 1)
        for i in range(self.__delta):
            for j in range(self.__M):
                temp[j] = self.__deficiency_pars[i] * self.__W_nullspace[i][j]
            rhs += temp
        # creating H(c,\alpha,k) with rows that might
        # be linearly dependent
        temp_vec = self.__cgraph.get_a() * self.__cgraph.get_psi() - rhs
        temp_vec2 = self.__cgraph.get_a() * psi_symbol_vec - rhs

        # creating a matrix of coefficients for the variables
        # psi_1, ..., psi_M, alpha_1, ... alpha_delta
        variables = [sympy.Symbol('psi' + str(i)) for i in range(self.__M)] + self.__deficiency_pars
        a, _ = sympy.linear_eq_to_matrix(temp_vec2, variables)

        reordered_indices = self.__create_fixed_free_pars_and_reordered_ind(temp_vec)

        temp_mat_2 = sympy.zeros(self.__M, self.__M + self.__delta)
        for i in range(self.__M):
            temp_mat_2[i, :] = a[reordered_indices[i], :]

        # gives vals which is the rows of temp_mat that are
        # independent from the others this allows us
        # to create H(c, \alpha, k)
        _, temp_vals = temp_mat_2.T.rref()
        vals = [reordered_indices[i] for i in temp_vals]

        # preallocating the H vector
        self.__H = sympy.zeros(self.__M - self.__ell, 1)

        # Filling the H vector with linearly independent rows
        for i in range(len(vals)):
            self.__H[i] = temp_vec[vals[i]]

        end = time.process_time()
        print("Elapsed time for creating Equilibrium Manifold: " + str(end - start))
        print("")

    def __create_fixed_free_pars_and_reordered_ind(self, temp_vec):

        # determining the different combinations of concentrations
        # present in the independent variables once the deficiency
        # parameters are chosen to be independent
        leng = self.__M - self.__ell - self.__delta
        comb = list(itertools.combinations(self.__concentration_pars, leng))

        # buidling the different possible independent variable sets
        indp_vars = []
        for i in comb:
            indp_vars.append(list(i)+self.__deficiency_pars)

        # finding the number of linear equations produced by a
        # given independent variable set
        counts = []
        for jj in range(len(indp_vars)):

            num_lin_entries = [self.__is_linear(temp_vec[j], indp_vars[jj][0:leng]) for j in
                               range(temp_vec.shape[0])].count(True)
            counts.append(num_lin_entries)

            # if all of the equations are linear stop,
            # prevents long run times
            if num_lin_entries == self.__M:
                break

        # picking the independent variable set that has the most
        # amount of linear equations
        chosen_index = counts.index(max(counts))

        self.__fixed_pars = indp_vars[chosen_index]
        self.__free_pars = [i for i in self.__deficiency_pars + self.__concentration_pars if i not in self.__fixed_pars]

        # rearranging A s.t. the linear equations are in the top
        # rows, this is for convenience and easier solutions for
        # the independent variables
        out = [self.__is_linear(temp_vec[j], indp_vars[chosen_index]) for j in range(temp_vec.shape[0])]
        reordered_indices = [i for i, x in enumerate(out) if x] + [i for i, x in enumerate(out) if not x]

        return reordered_indices

    # routine that determines if a sympy expression is jointly
    # linear with respect to a given set of variables. This
    # test is conducted by seeing if the second order derivatives
    # are zero.
    def __is_linear(self, expr, variables):

        for x in variables:
            for y in variables:
                try: 
                    if not sympy.Eq(sympy.diff(expr, x, y), 0):
                        return False
                except TypeError:
                    return False
        return True

    def __create_g_matrix(self):

        # creating the matrix DCH which is the jacobian of
        # the vector H with respect to the concentration vector
        self.__DCH = self.__H.jacobian(sympy.Matrix(self.__concentration_pars))

        # creation of the matrix dah which is the jacobian of
        # the vector H with respect to the deficiency param vector
        dah = self.__H.jacobian(sympy.Matrix(self.__deficiency_pars))
        # creation of the matrix daw which is the jacobian of
        # the vector W with respect to the deficiency param vector
        # However, as given in page 11, D_\alpha W = 0, thus
        # just a zero matrix of size \lambda by \delta
        daw = sympy.zeros(self.__lambda, self.__delta)
        # creating the upper half of the matrix G i.e. [DCH dah]
        g_upper = self.__DCH.row_join(dah)
        # creating the lower half of the matrix G i.e. [DCW daw]
        # Note that D_c W = B^T
        g_lower = self.__cgraph.get_b().row_join(daw)
        # putting the upper and lower half of the matrix together
        # this forms the full G(c, \alpha, k) matrix
        self.__G = g_upper.col_join(g_lower)

    def __create_symbolic_objective_fun(self):
        # computing the simplified version of the objective
        # function defined as: det(G(c, \alpha, k))^2
        self.__symbolic_objective_fun = (self.__G.det(method='lu'))**2

    def __create_concentration_values(self):

        print("Solving for species' concentrations ...")
        start = time.process_time()
        # Putting the concentrations in terms of the kinetic
        # constants and deficiency parameters using the H
        # vector of the equilibrium manifold
        temp_solution_tuple = sympy.solve(self.__H, self.__fixed_pars, dict=True)

        if not temp_solution_tuple:
            raise Exception("A solution for the fixed parameters could not be found!") 

        # construction concentrations if only one solution is found
        if isinstance(temp_solution_tuple, dict):
            self.__concentration_vals = []
            for i in self.__concentration_pars:
                if i in temp_solution_tuple:
                    self.__concentration_vals.append(temp_solution_tuple[i])
                else:
                    self.__concentration_vals.append(i)
        # multiple solutions found
        else:
            solution_list = []
            for i in temp_solution_tuple:
                temp = []
                for j in self.__concentration_pars:
                    if j in i:
                        temp.append(i[j])
                    else:
                        temp.append(j)
                solution_list.append(temp)
            self.__concentration_vals = self.__pick_solution_set(solution_list)

        end = time.process_time()
        print("Elapsed time for finding species' concentrations: " + str(end - start))
        print("")

    # chose solution set that is most likely to produce
    # positive concentrations
    def __pick_solution_set(self, solution_list):
        positivity = []
        for i in solution_list:
            temp = []
            for j in i:
                temp.append(j.is_positive)
            positivity.append(temp)

        verdict = []
        for i in positivity:
            if False not in i:
                if all(i):
                    # positive concentrations achieved
                    verdict.append("P")
                elif True in i:
                    # positive concentrations and Nones
                    verdict.append("PU")
                else:
                    # All entries are None
                    verdict.append("U")
            else:
                # negative concentration given
                verdict.append("N")

        if "P" in verdict:
            indx = verdict.index("P")
            choice = solution_list[indx]
        elif "PU" in verdict:
            indx = verdict.index("PU")
            choice = solution_list[indx]
        elif "U" in verdict:
            indx = verdict.index("U")
            choice = solution_list[indx]
        else:
            print("Solution chosen produces all negative concentrations!")
            sys.exit()

        return choice 

    def __create_decision_vector_x(self):
        # if it is a proper/over-dimensioned network let
        # xvec = (k_1 , ... k_R, alpha_1, ... alpha_lambda)
        # else let
        # xvec = (k_1 , ... k_R, alpha_1, ... alpha_delta,
        # c_1, ..., c_(lambda - delta))
        self.__decision_vector_x = self.__reaction_pars + self.__free_pars
        
        self.__important_info = "The network is " + self.__classification[0] + "\n"

        self.__d_len = len(self.__decision_vector_x)

    def __create_concentration_bounds_species(self):
        self.__concentration_bounds_species = [i for i in self.__concentration_pars
                                               if i not in self.__decision_vector_x]

    def __create_concentration_lambda_fun(self):
        self.__concentration_funs = []
        for i in range(self.__N):
            self.__concentration_funs += [sympy.utilities.lambdify(self.__decision_vector_x,
                                                                   self.__concentration_vals[i])]

    def __create_objective_fun__lambda_fun(self):
        self.__objective_fun_params = self.__reaction_pars + self.__concentration_pars
        self.__lambda_objective_fun = sympy.utilities.lambdify(self.__objective_fun_params,
                                                               self.__symbolic_objective_fun)

    def __create_g_matrix_lambda_fun(self):
        self.__lambda_G_matrix = sympy.utilities.lambdify(self.__objective_fun_params, self.__G)
        
    def __create_dch_matrix_lambda_fun(self):
        self.__lambda_DCH_matrix = sympy.utilities.lambdify(self.__objective_fun_params, self.__DCH)

    # getters
    def get_w_matrix(self):
        """
        Returns SymPy matrix :math:`[Y, \Lambda^T]^T`, which we call the W matrix.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_mass_conservation_approach()
            Creating Equilibrium Manifold ...
            Elapsed time for creating Equilibrium Manifold: 0.7020819999999999
            Solving for species' concentrations ...
            Elapsed time for finding species' concentrations: 0.888776
        >>> sympy.pprint(approach.get_w_matrix())
            ⎡1  0  0  0  0  1  1  0  0⎤
            ⎢                         ⎥
            ⎢1  0  1  0  0  0  0  0  0⎥
            ⎢                         ⎥
            ⎢0  1  0  0  0  0  0  0  0⎥
            ⎢                         ⎥
            ⎢0  0  1  1  0  0  1  0  2⎥
            ⎢                         ⎥
            ⎢0  0  0  1  0  1  0  0  0⎥
            ⎢                         ⎥
            ⎢0  0  0  0  0  0  0  1  0⎥
            ⎢                         ⎥
            ⎢0  0  0  0  1  0  0  0  0⎥
            ⎢                         ⎥
            ⎢1  1  1  0  0  0  0  0  0⎥
            ⎢                         ⎥
            ⎢0  0  0  1  1  1  0  0  0⎥
            ⎢                         ⎥
            ⎣0  0  0  0  0  0  1  1  1⎦
        """
        return self.__W

    def get_w_nullspace(self):
        """
        Returns a list of SymPy column vectors representing :math:`Null([Y, \Lambda^T]^T)`.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_mass_conservation_approach()
            Creating Equilibrium Manifold ...
            Elapsed time for creating Equilibrium Manifold: 0.7020819999999999
            Solving for species' concentrations ...
            Elapsed time for finding species' concentrations: 0.888776
        >>> sympy.pprint(approach.get_w_nullspace())
            ⎡⎡-1⎤  ⎡1 ⎤⎤
            ⎢⎢  ⎥  ⎢  ⎥⎥
            ⎢⎢0 ⎥  ⎢0 ⎥⎥
            ⎢⎢  ⎥  ⎢  ⎥⎥
            ⎢⎢1 ⎥  ⎢-1⎥⎥
            ⎢⎢  ⎥  ⎢  ⎥⎥
            ⎢⎢-1⎥  ⎢0 ⎥⎥
            ⎢⎢  ⎥  ⎢  ⎥⎥
            ⎢⎢0 ⎥, ⎢0 ⎥⎥
            ⎢⎢  ⎥  ⎢  ⎥⎥
            ⎢⎢1 ⎥  ⎢0 ⎥⎥
            ⎢⎢  ⎥  ⎢  ⎥⎥
            ⎢⎢0 ⎥  ⎢-1⎥⎥
            ⎢⎢  ⎥  ⎢  ⎥⎥
            ⎢⎢0 ⎥  ⎢0 ⎥⎥
            ⎢⎢  ⎥  ⎢  ⎥⎥
            ⎣⎣0 ⎦  ⎣1 ⎦⎦
        """
        return self.__W_nullspace

    def get_h_vector(self):
        """
        Returns a SymPy matrix representing the equilibrium manifold.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_mass_conservation_approach()
            Creating Equilibrium Manifold ...
            Elapsed time for creating Equilibrium Manifold: 0.7020819999999999
            Solving for species' concentrations ...
            Elapsed time for finding species' concentrations: 0.888776
        >>> sympy.pprint(approach.get_h_vector())
            ⎡a₁ - a₂ - re₁⋅s₁⋅s₂ + re1r⋅s₃⎤
            ⎢                             ⎥
            ⎢re₁⋅s₁⋅s₂ + s₃⋅(-re1r - re₂) ⎥
            ⎢                             ⎥
            ⎢  a₁ - re₃⋅s₆⋅s₇ + re3r⋅s₁₆  ⎥
            ⎢                             ⎥
            ⎢re₃⋅s₆⋅s₇ + s₁₆⋅(-re3r - re₄)⎥
            ⎢                             ⎥
            ⎢  a₂ - re₆⋅s₁⋅s₆ + re6r⋅s₁₅  ⎥
            ⎢                             ⎥
            ⎣re₆⋅s₁⋅s₆ + s₁₅⋅(-re6r - re₈)⎦
        """
        return self.__H

    def get_g_matrix(self):
        """
        Returns a SymPy matrix representing the G matrix of the defined optimization problem.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_mass_conservation_approach()
            Creating Equilibrium Manifold ...
            Elapsed time for creating Equilibrium Manifold: 0.7020819999999999
            Solving for species' concentrations ...
            Elapsed time for finding species' concentrations: 0.888776
        >>> sympy.pprint(approach.get_g_matrix())
            ⎡-re₁⋅s₂  -re₁⋅s₁     re1r         0        0          0            0       1  -1⎤
            ⎢                                                                                ⎥
            ⎢re₁⋅s₂   re₁⋅s₁   -re1r - re₂     0        0          0            0       0  0 ⎥
            ⎢                                                                                ⎥
            ⎢   0        0          0       -re₃⋅s₇  -re₃⋅s₆       0          re3r      1  0 ⎥
            ⎢                                                                                ⎥
            ⎢   0        0          0       re₃⋅s₇   re₃⋅s₆        0       -re3r - re₄  0  0 ⎥
            ⎢                                                                                ⎥
            ⎢-re₆⋅s₆     0          0       -re₆⋅s₁     0        re6r           0       0  1 ⎥
            ⎢                                                                                ⎥
            ⎢re₆⋅s₆      0          0       re₆⋅s₁      0     -re6r - re₈       0       0  0 ⎥
            ⎢                                                                                ⎥
            ⎢   0        0          0          0       1.0         0           1.0      0  0 ⎥
            ⎢                                                                                ⎥
            ⎢   0       1.0        1.0         0        0          0            0       0  0 ⎥
            ⎢                                                                                ⎥
            ⎣  1.0       0         1.0        1.0       0         2.0          1.0      0  0 ⎦
        """
        return self.__G

    def get_dch_matrix(self):
        """
        Returns a SymPy matrix representing the Jacobian of the equilibrium manifold with respect to the species.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_mass_conservation_approach()
            Creating Equilibrium Manifold ...
            Elapsed time for creating Equilibrium Manifold: 0.7020819999999999
            Solving for species' concentrations ...
            Elapsed time for finding species' concentrations: 0.888776
        >>> sympy.pprint(approach.get_dch_matrix())
            ⎡-re₁⋅s₂  -re₁⋅s₁     re1r         0        0          0            0     ⎤
            ⎢                                                                         ⎥
            ⎢re₁⋅s₂   re₁⋅s₁   -re1r - re₂     0        0          0            0     ⎥
            ⎢                                                                         ⎥
            ⎢   0        0          0       -re₃⋅s₇  -re₃⋅s₆       0          re3r    ⎥
            ⎢                                                                         ⎥
            ⎢   0        0          0       re₃⋅s₇   re₃⋅s₆        0       -re3r - re₄⎥
            ⎢                                                                         ⎥
            ⎢-re₆⋅s₆     0          0       -re₆⋅s₁     0        re6r           0     ⎥
            ⎢                                                                         ⎥
            ⎣re₆⋅s₆      0          0       re₆⋅s₁      0     -re6r - re₈       0     ⎦
        """
        return self.__DCH

    def get_lambda_g_matrix(self):
        """
        Returns a lambda function representation of the G matrix. Here the arguments of the lambda function are given
        by the values provided by :func:`crnt4sbml.MassConservationApproach.get_objective_fun_params`.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_mass_conservation_approach()
            Creating Equilibrium Manifold ...
            Elapsed time for creating Equilibrium Manifold: 0.7020819999999999
            Solving for species' concentrations ...
            Elapsed time for finding species' concentrations: 0.888776
        >>> print(approach.get_lambda_g_matrix())
            <function _lambdifygenerated at 0x13248ac80>
        """
        return self.__lambda_G_matrix

    def get_lambda_dch_matrix(self):
        """
        Returns a lambda function representation of the Jacobian of the equilibrium manifold matrix. Here the
        arguments of the lambda function are given by the values provided by
        :func:`crnt4sbml.MassConservationApproach.get_objective_fun_params`.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_mass_conservation_approach()
            Creating Equilibrium Manifold ...
            Elapsed time for creating Equilibrium Manifold: 0.7020819999999999
            Solving for species' concentrations ...
            Elapsed time for finding species' concentrations: 0.888776
        >>> print(approach.get_lambda_dch_matrix())
            <function _lambdifygenerated at 0x131a06ea0>
        """
        return self.__lambda_DCH_matrix

    def get_symbolic_objective_fun(self):
        """
        Returns SymPy expression for the objective function of the optimization problem. This is the determinant of the
        G matrix squared.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_mass_conservation_approach()
            Creating Equilibrium Manifold ...
            Elapsed time for creating Equilibrium Manifold: 0.7020819999999999
            Solving for species' concentrations ...
            Elapsed time for finding species' concentrations: 0.888776
        >>> print(approach.get_symbolic_objective_fun())
            1.0*re1**2*re2**2*re3**2*re4**2*re6**2*re8**2*s1**2*s6**2*s7**2*((1.0*s2/s7 - 1.0*s2*(-re3r - re4)/
            (re3*s6*s7))/re4 + (1.0 + 1.0*re1r/(re1*s1))/re2 + 1.0/(re1*s1))**2*(-((1.0*s6*(-1.0*s1/s6 + 1.0)/s7 + 1.0
            - (-re3r - re4)*(-1.0*s1/s6 + 1.0)/(re3*s7))/re4 + 1.0/re2)*(-1.0*re6r*s2/(re6*re8*s1*s6) - 1.0*s2*(1 +
            re6*s6/(re1*s2))/(re6*s1*s6) - (1.0 + 1.0*re1r/(re1*s1))/re2)/((1.0*s2/s7 - 1.0*s2*(-re3r - re4)/
            (re3*s6*s7))/re4 + (1.0 + 1.0*re1r/(re1*s1))/re2 + 1.0/(re1*s1)) + (2.0 + 1.0*re6r/(re6*s6))/re8 + 1.0*(1 +
            re6*s6/(re1*s2))/(re6*s6) - 1.0/re2 - 1.0/(re1*s2))**2
        """
        return self.__symbolic_objective_fun

    def get_lambda_objective_fun(self):
        """
        Returns a lambda function representation of the objective function of the optimization problem. Here the
        arguments of the lambda function are given by the values provided by
        :func:`crnt4sbml.MassConservationApproach.get_objective_fun_params`.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_mass_conservation_approach()
            Creating Equilibrium Manifold ...
            Elapsed time for creating Equilibrium Manifold: 0.7020819999999999
            Solving for species' concentrations ...
            Elapsed time for finding species' concentrations: 0.888776
        >>> print(approach.get_lambda_objective_fun())
            <function _lambdifygenerated at 0x12f6f7ea0>
        """
        return self.__lambda_objective_fun

    def get_concentration_vals(self):
        """
        Returns a list of SymPy expressions representing the species in terms of those variables present in the decision
        vector. The order is that established in :func:`crnt4sbml.Cgraph.get_species`. Note that if only a single
        species is provided as an element in the list, this means the species is a free variable.

        See also
        ---------
        crnt4sbml.MassConservationApproach.get_concentration_solutions


        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_mass_conservation_approach()
            Creating Equilibrium Manifold ...
            Elapsed time for creating Equilibrium Manifold: 0.7020819999999999
            Solving for species' concentrations ...
            Elapsed time for finding species' concentrations: 0.888776
        >>> print(approach.get_concentration_vals())
            [re4*s16*(re1r*re6r + re1r*re8 + re2*re6r + re2*re8)/(re1*re2*re6r*s2 + re1*re2*re8*s2 + re1r*re6*re8*s6 +
            re2*re6*re8*s6), s2, re1*re4*s16*s2*(re6r + re8)/(re1*re2*re6r*s2 + re1*re2*re8*s2 + re1r*re6*re8*s6 +
            re2*re6*re8*s6), s6, s16*(re3r + re4)/(re3*s6), re4*re6*s16*s6*(re1r + re2)/(re1*re2*re6r*s2 +
            re1*re2*re8*s2 + re1r*re6*re8*s6 + re2*re6*re8*s6), s16]
        """
        return self.__concentration_vals

    def get_decision_vector(self):
        """
        Returns a list of SymPy variables that represent the decision vector of the optimization problem.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_mass_conservation_approach()
            Creating Equilibrium Manifold ...
            Elapsed time for creating Equilibrium Manifold: 0.7020819999999999
            Solving for species' concentrations ...
            Elapsed time for finding species' concentrations: 0.888776
        >>> print(approach.get_decision_vector())
            [re1, re1r, re2, re3, re3r, re4, re6, re6r, re8, s2, s6, s16]
        """
        return self.__decision_vector_x

    def get_concentration_bounds_species(self):
        """
        Returns a list of SymPy variables that represents the order of species for the concentration bounds provided
        to :func:`crnt4sbml.MassConservationApproach.run_optimization`.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_mass_conservation_approach()
            Creating Equilibrium Manifold ...
            Elapsed time for creating Equilibrium Manifold: 0.7020819999999999
            Solving for species' concentrations ...
            Elapsed time for finding species' concentrations: 0.888776
        >>> print(approach.get_concentration_bounds_species())
            [s1, s3, s7, s15]
        """
        return self.__concentration_bounds_species

    def get_concentration_funs(self):
        """
        Returns a list of lambda functions representing each of the species. Here the species are those expressions
        provided by :func:`crnt4sbml.MassConservationApproach.get_concentration_vals` where the arguments of each
        lambda function is provided by :func:`crnt4sbml.MassConservationApproach.get_decision_vector`.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_mass_conservation_approach()
            Creating Equilibrium Manifold ...
            Elapsed time for creating Equilibrium Manifold: 0.7020819999999999
            Solving for species' concentrations ...
            Elapsed time for finding species' concentrations: 0.888776
        >>> print(approach.get_concentration_funs())
            [<function _lambdifygenerated at 0x12b673f28>, <function _lambdifygenerated at 0x12b7a6ea0>,
            <function _lambdifygenerated at 0x12b7a6d08>, <function _lambdifygenerated at 0x12b7a67b8>,
            <function _lambdifygenerated at 0x12b7a6d90>, <function _lambdifygenerated at 0x12b7a6488>,
            <function _lambdifygenerated at 0x12b7a6950>]
        """
        return self.__concentration_funs

    def get_objective_fun_params(self):
        """
        Returns a list of SymPy variables that represent those variables that may be contained in the G matrix, Jacobian
        of the equilibrium manifold with respect to the species, or objective function.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_mass_conservation_approach()
            Creating Equilibrium Manifold ...
            Elapsed time for creating Equilibrium Manifold: 0.7020819999999999
            Solving for species' concentrations ...
            Elapsed time for finding species' concentrations: 0.888776
        >>> print(approach.get_objective_fun_params())
            [re1, re1r, re2, re3, re3r, re4, re6, re6r, re8, s1, s2, s3, s6, s7, s15, s16]
        """
        return self.__objective_fun_params

    def get_conservation_laws(self):
        """
        Returns a string representation of the conservation laws. Here the values on the left hand side of each equation
        are the constants of the conservation laws.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_mass_conservation_approach()
            Creating Equilibrium Manifold ...
            Elapsed time for creating Equilibrium Manifold: 0.7020819999999999
            Solving for species' concentrations ...
            Elapsed time for finding species' concentrations: 0.888776
        >>> print(approach.get_conservation_laws())
            C1 = 1.0*s16 + 1.0*s7
            C2 = 1.0*s2 + 1.0*s3
            C3 = 1.0*s1 + 2.0*s15 + 1.0*s16 + 1.0*s3 + 1.0*s6
        """
        rhs = self.__cgraph.get_b() * sympy.Matrix([self.__concentration_pars]).T
        laws = ""
        for i in range(rhs.shape[0]):
            laws += 'C' + str(i+1) + ' = ' + str(rhs[i]) + '\n'
        
        return laws

    def get_concentration_solutions(self):
        """
        Returns a more readable string representation of the species defined in terms of the decision vector.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> approach = network.get_mass_conservation_approach()
            Creating Equilibrium Manifold ...
            Elapsed time for creating Equilibrium Manifold: 0.7020819999999999
            Solving for species' concentrations ...
            Elapsed time for finding species' concentrations: 0.888776
        >>> print(approach.get_concentration_solutions())
            s1 = re4*s16*(re1r*re6r + re1r*re8 + re2*re6r + re2*re8)/(re1*re2*re6r*s2 + re1*re2*re8*s2 + re1r*re6*re8*s6 + re2*re6*re8*s6)
            s2 = s2
            s3 = re1*re4*s16*s2*(re6r + re8)/(re1*re2*re6r*s2 + re1*re2*re8*s2 + re1r*re6*re8*s6 + re2*re6*re8*s6)
            s6 = s6
            s7 = s16*(re3r + re4)/(re3*s6)
            s15 = re4*re6*s16*s6*(re1r + re2)/(re1*re2*re6r*s2 + re1*re2*re8*s2 + re1r*re6*re8*s6 + re2*re6*re8*s6)
            s16 = s16
        """
        sols = ""
        for i in range(self.__N):
            sols += self.__species[i] + ' = ' + str(self.__concentration_vals[i]) + '\n'            

        return sols 

    def run_optimization(self, bounds=None, iterations=10, sys_min_val=numpy.finfo(float).eps, seed=0, print_flag=False,
                         numpy_dtype=numpy.float64, concentration_bounds=None):
        """
        Function for running the optimization problem for the mass conservation approach.

        Parameters
        -----------
            bounds: list of tuples
                A list defining the lower and upper bounds for each variable in the decision vector. Here the reactions
                are allowed to be set to a single value.
            iterations: int
                The number of iterations to run the feasible point method.
            sys_min_val: float
                The value that should be considered zero for the optimization problem.
            seed: int
                Seed for the random number generator. None should be used if a random generation is desired.
            print_flag: bool
                Should be set to True if the user wants the objective function values found in the optimization problem
                and False otherwise.
            numpy_dtype:
                The numpy data type used within the optimization routine. All variables in the optimization routine will
                be converted to this data type.
            concentration_bounds: list of tuples
                A list defining the lower and upper bounds for those species' concentrations not in the decision vector.
                The user is not allowed to set the species' concentration to a single value. See also:
                :func:`crnt4sbml.MassConservationApproach.get_concentration_bounds_species`.
        Returns
        --------
        params_for_global_min: list of numpy arrays
            A list of numpy arrays that correspond to the decision vectors of the problem.
        obj_fun_val_for_params: list of floats
            A list of objective function values produced by the corresponding decision vectors in params_for_global_min.

        Examples
        ---------
        See :ref:`quickstart-deficiency-label` and :ref:`my-deficiency-label`.
        """
        self.__numpy_dtype = numpy_dtype
        temp_c = numpy.zeros(self.__N, dtype=self.__numpy_dtype)

        # testing to see if there are any equalities in bounds
        equality_bounds_indices = []
        for i in range(len(bounds)):
            if not isinstance(bounds[i], tuple):
                equality_bounds_indices.append(i)

        # recasting user provided input to numpy_dtype
        for i in range(len(bounds)):
            bounds[i] = self.__numpy_dtype(bounds[i])
            
        for i in range(len(concentration_bounds)):
            concentration_bounds[i] = self.__numpy_dtype(concentration_bounds[i])

        if len(concentration_bounds) != len(self.__concentration_bounds_species):
            print("Concentration bounds is the incorrect length!")
            sys.exit()

        full_concentration_bounds = []
        for i in range(self.__N):
            if self.__concentration_pars[i] in self.__decision_vector_x:
                indx = self.__decision_vector_x.index(self.__concentration_pars[i])
                full_concentration_bounds.append(bounds[indx])
            else:
                indx = self.__concentration_bounds_species.index(self.__concentration_pars[i])
                full_concentration_bounds.append(concentration_bounds[indx])

        sys_min_val = self.__numpy_dtype(sys_min_val)

        params_for_global_min, obj_fun_val_for_params, self.__important_info = BistabilityFinder.run_optimization(
            bounds, iterations, sys_min_val, temp_c, self.__penalty_objective_func, self.__feasible_point_check,
            self.__objective_function_to_optimize, self.__final_constraint_check, seed, equality_bounds_indices,
            print_flag, numpy_dtype, full_concentration_bounds)

        return params_for_global_min, obj_fun_val_for_params

    def run_continuity_analysis(self, species=None, parameters=None, dir_path="./num_cont_graphs",
                                print_lbls_flag=False, auto_parameters=None):
        """
        Function for running the numerical continuation and bistability analysis portions of the mass conservation
        approach.

        Parameters
        ------------
            species: string
                A string stating the species that is the y-axis of the bifurcation diagram.
            parameters: list of numpy arrays
                A list of numpy arrays corresponding to the decision vectors that produce a small objective function
                value.
            dir_path: string
                A string stating the path where the bifurcation diagrams should be saved.
            print_lbls_flag: bool
                If True the routine will print the special points found by AUTO 2000 and False will not print any
                special points.
            auto_parameters: dict
                Dictionary defining the parameters for the AUTO 2000 run. Please note that one should **not** set
                'SBML' or 'ScanDirection' in these parameters as these are automatically assigned. It is absolutely
                necessary to set PrincipalContinuationParameter in this dictionary. For more information on these
                parameters refer to :download:`AUTO parameters <../auto2000_input.pdf>`. 'NMX' will default to
                10000 and 'ITMX' to 100.
        Returns
        ---------
            multistable_param_ind: list of integers
                A list of those indices in 'parameters' that produce multistable plots.

        Example
        ---------
        See :ref:`quickstart-deficiency-label` and :ref:`my-deficiency-label`.
        """
        # setting default values for AUTO
        if 'NMX' not in auto_parameters.keys():
            auto_parameters['NMX'] = 10000

        if 'ITMX' not in auto_parameters.keys():
            auto_parameters['ITMX'] = 100

        # making the directory if it doesn't exist
        if not os.path.isdir(dir_path):
            os.mkdir(dir_path)

        species_num = self.__species.index(species) + 1

        species_y = str(self.__concentration_pars[species_num-1]) 

        multistable_param_ind, important_info = BistabilityFinder.run_continuity_analysis(species_num, parameters,
                                                                                          self.__initialize_ant_string,
                                                                                          self.__finalize_ant_string,
                                                                                          species_y, dir_path,
                                                                                          print_lbls_flag,
                                                                                          auto_parameters)

        self.__important_info += important_info

        return multistable_param_ind

    def run_greedy_continuity_analysis(self, species=None, parameters=None, dir_path="./num_cont_graphs",
                                       print_lbls_flag=False, auto_parameters=None):
        """
        Function for running the greedy numerical continuation and bistability analysis portions of the mass conservation
        approach. This routine uses the initial value of the principal continuation parameter to construct AUTO
        parameters and then tests varying fixed step sizes for the continuation problem. Note that this routine may
        produce jagged or missing sections in the plots provided. To produce better plots one should use the information
        provided by this routine to run :func:`crnt4sbml.MassConservationApproach.run_continuity_analysis`.

        Parameters
        ------------
            species: string
                A string stating the species that is the y-axis of the bifurcation diagram.
            parameters: list of numpy arrays
                A list of numpy arrays corresponding to the decision vectors that produce a small objective function
                value.
            dir_path: string
                A string stating the path where the bifurcation diagrams should be saved.
            print_lbls_flag: bool
                If True the routine will print the special points found by AUTO 2000 and False will not print any
                special points.
            auto_parameters: dict
                Dictionary defining the parameters for the AUTO 2000 run. Please note that only the
                PrincipalContinuationParameter in this dictionary should be defined, no other AUTO parameters should
                be set. For more information on these parameters refer to :download:`AUTO parameters <../auto2000_input.pdf>`.
        Returns
        ---------
            multistable_param_ind: list of integers
                A list of those indices in 'parameters' that produce multistable plots.

        Example
        ---------
        See :ref:`my-deficiency-label`.
        """
        # setting default values for AUTO
        if 'NMX' not in auto_parameters.keys():
            auto_parameters['NMX'] = 10000

        if 'ITMX' not in auto_parameters.keys():
            auto_parameters['ITMX'] = 100

        # making the directory if it doesn't exist
        if not os.path.isdir(dir_path):
            os.mkdir(dir_path)

        species_num = self.__species.index(species) + 1

        species_y = str(self.__concentration_pars[species_num-1])

        multistable_param_ind, important_info = BistabilityFinder.run_greedy_continuity_analysis\
            (species_num, parameters, self.__initialize_ant_string, self.__finalize_ant_string, species_y, dir_path,
             print_lbls_flag, auto_parameters)

        self.__important_info += important_info

        return multistable_param_ind

    def generate_report(self):
        """
        Prints out helpful details constructed by :func:`crnt4sbml.MassConservationApproach.run_optimization` and
        :func:`crnt4sbml.MassConservationApproach.run_continuity_analysis`.

        Example
        --------
        See :ref:`quickstart-deficiency-label` and :ref:`my-deficiency-label`.
        """
        print(self.__important_info)

    def __initialize_ant_string(self, species_num, pcp_x):
        y = self.__cgraph.get_y()
        a = self.__cgraph.get_a()
        bt = self.__cgraph.get_b()
        psi = self.__cgraph.get_psi()

        # forming ya matrix
        ya = y*a

        # finding how many rows are indep in ya
        _, vals = ya.T.rref()
        num_indp_eqns = len(vals)
        num_dep_eqns = ya.shape[0] - num_indp_eqns

        # getting dimensions of bt
        bt_rows = bt.shape[0]
        bt_cols = bt.shape[1]

        bt_nonzero_ind = []
        for i in range(bt_rows):
            bt_nonzero_ind.append([j for j in range(bt_cols) if bt[i, j] != 0 and j != species_num-1])

        chosen_indp_indices, chosen_dep_indices = self.__get_indp_dep_species_indices(bt_nonzero_ind, num_dep_eqns,
                                                                                      num_indp_eqns, ya)

        replacements, ind_spec_conc_temp, indp_odes_temp = self.__construct_important_variables(chosen_indp_indices,
                                                                                                chosen_dep_indices, ya,
                                                                                                psi, bt)

        ode_str = self.__create_ode_str(replacements, ind_spec_conc_temp, indp_odes_temp, species_num)

        return ode_str, pcp_x

    def __get_indp_dep_species_indices(self, bt_nonzero_ind, num_dep_eqns, num_indp_eqns, ya):
        # getting all combinations of the list indices
        possible_dep_species = list(itertools.product(*bt_nonzero_ind))

        removed_entries = []
        # remove tuples that have duplicate entries
        for i in range(len(possible_dep_species)):
            if len(set(possible_dep_species[i])) != num_dep_eqns:
                removed_entries.append(i)
        for index in sorted(removed_entries, reverse=True):
            del possible_dep_species[index]

        # get corresponding possible dependent species
        possible_indp_species = [] 
        species_ind = [i for i in range(len(self.__concentration_pars))]

        for i in possible_dep_species:
            possible_indp_species.append([j for j in species_ind if j not in i])

        # using YA to pick one of the possible indices
        chosen_indp_indices = []
        chosen_dep_indices = []
        for i in range(len(possible_indp_species)):
            _, vals = ya[possible_indp_species[i], :].T.rref()

            if len(vals) == num_indp_eqns: 
                chosen_indp_indices = possible_indp_species[i]
                chosen_dep_indices = possible_dep_species[i]
                break

        return chosen_indp_indices, chosen_dep_indices

    def __construct_important_variables(self, chosen_indp_indices, chosen_dep_indices, ya, psi, bt):
        # getting independent concentrations
        ind_spec_conc_temp = [self.__concentration_pars[i] for i in chosen_indp_indices]

        # getting dependent concentrations
        dep_spec_conc = [self.__concentration_pars[i] for i in chosen_dep_indices]

        # constructing the independent ODEs
        indp_odes_temp = ya[chosen_indp_indices, :] * psi

        # creating conservation laws string
        self.__cons_laws_sympy = bt*sympy.Matrix([self.__concentration_pars]).T

        # Lambda function of conservation laws
        self.__cons_laws_lamb = [sympy.utilities.lambdify(self.__concentration_pars, self.__cons_laws_sympy[i])
                                 for i in range(len(self.__cons_laws_sympy))]

        cons_laws_sympy_eq = [sympy.Eq(sympy.Symbol('C' + str(i+1), real=True), self.__cons_laws_sympy[i])
                              for i in range(len(self.__cons_laws_sympy))]

        dep_conc_in_laws = self.__dependent_species_concentrations(self.__cons_laws_sympy, dep_spec_conc)
        
        replacements = self.__find_dep_concentration_replacements(dep_conc_in_laws, self.__cons_laws_sympy,
                                                                  dep_spec_conc, cons_laws_sympy_eq)

        return replacements, ind_spec_conc_temp, indp_odes_temp

    def __create_ode_str(self, replacements, ind_spec_conc_temp, indp_odes_temp, species_num):
        # rearrange ind_spec_conc and indp_odes to make species of
        # interest be the first ODE
        indx_species_num = ind_spec_conc_temp.index(self.__concentration_pars[species_num-1])

        self.__ind_spec_conc = [ind_spec_conc_temp[indx_species_num]]
        for i in ind_spec_conc_temp:
            if i != self.__concentration_pars[species_num-1]:
                self.__ind_spec_conc.append(i)

        indp_odes = sympy.zeros(indp_odes_temp.shape[0], indp_odes_temp.shape[1])
        indp_odes[0] = indp_odes_temp[indx_species_num]
        count = 1
        for i in range(indp_odes_temp.shape[0]):
            if i != indx_species_num:
                indp_odes[count] = indp_odes_temp[i]
                count += 1

        # bulding ODE string in Antimony format
        ode_str = self.__building_ode_str(replacements, self.__ind_spec_conc, indp_odes)

        return ode_str
        
    def __finalize_ant_string(self, x, ode_str):
        concentration_vals = [self.__concentration_funs[j](*tuple(x)) for j in range(self.__N)]

        kinetic_vals = [x[i] for i in range(self.__R)] 

        antstr = self.__initialize_variables_in_antimony_string(self.__cons_laws_sympy, ode_str,
                                                                self.__cons_laws_lamb, concentration_vals, kinetic_vals,
                                                                self.__reaction_pars)

        return antstr

    def __final_constraint_check(self, x_initial, penalty_bounds, sys_min_val, equality_bounds_indices,
                                 concentration_bounds):
        non_equality_bounds_indices = [i for i in range(len(penalty_bounds)) if i not in equality_bounds_indices]

        x = numpy.zeros(len(penalty_bounds), dtype=self.__numpy_dtype)
        for j in equality_bounds_indices:              
            x[j] = penalty_bounds[j]              
        count = 0                                      
        for j in non_equality_bounds_indices:          
            x[j] = x_initial[count]                
            count += 1 

        # concentration > 0 check
        con = numpy.asarray([self.__concentration_funs[j](*tuple(x)) for j in range(self.__N)],
                            dtype=self.__numpy_dtype)
        con_temp = []
        for i in range(self.__N):
            con_temp.append(con[i] >= concentration_bounds[i][0] and con[i] <= concentration_bounds[i][1])
        concs_chk = numpy.all(con_temp)

        # boundary check
        test = []
        for j in non_equality_bounds_indices: 
            test.append(x[j] >= penalty_bounds[j][0] and x[j] <= penalty_bounds[j][1])
        boundry_chk = numpy.all(test)

        # rank(G) = N + delta - 1 check
        # xx = numpy.concatenate((x[0:self.__R],con),axis=None)

        # must convert xx to numpy.float64 because higher
        # is not supported in linalg
        # xx = numpy.float64(xx)

        # rank_G = numpy.linalg.matrix_rank(self.__lambda_G_matrix(*tuple(xx)))
        # rank_G_chk = rank_G == (self.__N + self.__delta - 1)

        # rank(DCH) = min(N,M-ell) check
        # rank_DCH = numpy.linalg.matrix_rank(self.__lambda_DCH_matrix(*tuple(xx)))

        # rank_DCH_chk = rank_DCH == min(self.__N,self.__M - self.__ell)

        if concs_chk and boundry_chk:  # and rank_G_chk and rank_DCH_chk:
            
            return [True, x]
        else:
            
            return [False, []]

    def __concentration_violation_fun(self, g, len_g):
        temp = numpy.zeros(len_g, dtype=self.__numpy_dtype)
        for i in range(len_g):
            temp[i] = numpy.maximum(self.__numpy_dtype(0.0), -g[i])**2
        return temp

    def __x_violation_fun(self, x, b, len_x):
        temp = numpy.zeros(len_x, dtype=self.__numpy_dtype)
        for i in range(len_x):
            temp[i] = numpy.maximum(self.__numpy_dtype(0.0), self.__numpy_dtype(b)-x[i])**2
        return temp

    def __penalty_objective_func(self, x_initial, temp_c, penalty_bounds, equality_bounds_indices, x,
                                 non_equality_bounds_indices, concentration_bounds):

        for j in equality_bounds_indices:
            x[j] = penalty_bounds[j]
        count = 0 
        for j in non_equality_bounds_indices:
            x[j] = x_initial[count]
            count += 1 

        # evaluating the concentrations first
        for i in range(self.__N):
            temp_val = self.__concentration_funs[i](*tuple(x))
            if numpy.iscomplex(temp_val): 
                temp_c = numpy.array([numpy.Inf for i in range(self.__N)], dtype=self.__numpy_dtype)
                break
            else:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", numpy.ComplexWarning)
                    temp_c[i] = temp_val 

        if numpy.all(numpy.isfinite(temp_c)):

            # obtaining the sum of the violation functions squared
            sumval = self.__numpy_dtype(0.0)
            for j in range(self.__N):
                sumval += numpy.maximum(self.__numpy_dtype(0.0), concentration_bounds[j][0]-temp_c[j])**2
                sumval += numpy.maximum(self.__numpy_dtype(0.0), temp_c[j] - concentration_bounds[j][1])**2

            sum0 = self.__numpy_dtype(0.0)
            for j in non_equality_bounds_indices: 
                sum0 += numpy.maximum(self.__numpy_dtype(0.0), penalty_bounds[j][0]-x[j])**2
                sum0 += numpy.maximum(self.__numpy_dtype(0.0), x[j] - penalty_bounds[j][1])**2
            sumval += sum0

            # obtaining the violation function values for
            # k's and concentrations in x
            # xx = numpy.concatenate((x[0:self.__R],x[self.__R + self.__alpha_end_ind:self.__d_len]),axis=None)
            # temp = self.__x_violation_fun(xx,self.__numpy_dtype(0.0),self.__R + (self.__d_len - (self.__R +
            # self.__alpha_end_ind)))
            # sumval += numpy.sum(temp)
            return sumval 

        else:
            return numpy.PINF

    def __feasible_point_check(self, x, result_fun, sys_min_val, equality_bounds_indices, non_equality_bounds_indices,
                               penalty_bounds, concentration_bounds):
        result_x = numpy.zeros(len(penalty_bounds), dtype=self.__numpy_dtype)

        for j in equality_bounds_indices:
            result_x[j] = penalty_bounds[j]
        count = 0
        for j in non_equality_bounds_indices:
            result_x[j] = x[count]
            count += 1

        # double checking the concentrations
        con = numpy.asarray([self.__concentration_funs[i](*tuple(result_x)) for i in range(self.__N)],
                            dtype=self.__numpy_dtype)
        con_temp = []
        for i in range(self.__N):
            con_temp.append(con[i] >= concentration_bounds[i][0] and con[i] <= concentration_bounds[i][1])
        concs_chk = numpy.all(con_temp)

        finite_chk = numpy.isfinite(con)
        if concs_chk and numpy.all(finite_chk):

            # putting the feasible points in x_candidates
            if abs(result_fun) <= sys_min_val and numpy.all(con > self.__numpy_dtype(0)):
                return True
            else:
                return False
        else:
            return False 
                
    def __objective_function_to_optimize(self, x_initial, temp_c, penalty_bounds, sys_min_val, equality_bounds_indices,
                                         x, non_equality_bounds_indices, concentration_bounds):
        for j in equality_bounds_indices:
            x[j] = penalty_bounds[j]
        count = 0
        for j in non_equality_bounds_indices:
            x[j] = x_initial[count]
            count += 1

        test = []
        for j in non_equality_bounds_indices:
            test.append(x[j] >= penalty_bounds[j][0] and x[j] <= penalty_bounds[j][1])
        boundry_chk = numpy.all(test)

        if boundry_chk:
            # calculating the concentration values
            for i in range(self.__N):
                temp_val = self.__concentration_funs[i](*tuple(x))
                if numpy.iscomplex(temp_val):
                    temp_c = numpy.array([numpy.Inf for i in range(self.__N)], dtype=self.__numpy_dtype)
                    break
                else:
                    with warnings.catch_warnings(): 
                        warnings.simplefilter("ignore", numpy.ComplexWarning)
                        temp_c[i] = temp_val 

            finite_chk = numpy.isfinite(temp_c)
            con_temp = []
            for i in range(self.__N):
                con_temp.append(temp_c[i] >= concentration_bounds[i][0] and temp_c[i] <= concentration_bounds[i][1])
            concs_chk = numpy.all(con_temp)
        # making sure our concentrations are finite
            if concs_chk and numpy.all(finite_chk): 
                temp = numpy.zeros(self.__N, dtype=self.__numpy_dtype)
                for i in range(self.__N):
                    temp[i] = numpy.maximum(self.__numpy_dtype(0.0), -temp_c[i])
                sumval = numpy.sum(temp)

                xx = numpy.concatenate((x[0:self.__R], temp_c), axis=None)
                return self.__lambda_objective_fun(*tuple(xx)) + sumval
            else:
                return numpy.PINF

        else:
            return numpy.PINF

    def __dependent_species_concentrations(self, cons_laws_sympy, dep_spec_conc):
        # finding those dep_spec_conc that occur in each conservation law
        dep_conc_in_laws = []
        for i in range(len(cons_laws_sympy)):
            temp = []
            for j in range(len(dep_spec_conc)):
                if cons_laws_sympy[i].count(dep_spec_conc[j]) > 0:
                    temp.append(dep_spec_conc[j])
            dep_conc_in_laws.append(temp)
        return dep_conc_in_laws

    def __is_list_empty(self, inlist):
        if isinstance(inlist, list):  # Is a list
            return all(map(self.__is_list_empty, inlist))
        return False  # Not a list

    def __find_dep_concentration_replacements(self, dep_conc_in_laws, cons_laws_sympy, dep_spec_conc,
                                              cons_laws_sympy_eq):
        replacements = []
        flag = True
        while flag:
            for i in range(len(cons_laws_sympy_eq)):
                if len(dep_conc_in_laws[i]) == 1:
                    temp = sympy.solve(cons_laws_sympy_eq[i], dep_conc_in_laws[i])
                    cons_laws_sympy = [cons_laws_sympy[j].subs(dep_conc_in_laws[i][0], temp[0])
                                       for j in range(len(cons_laws_sympy))]

                    cons_laws_sympy_eq = [sympy.Eq(sympy.Symbol('C' + str(i+1), real=True), cons_laws_sympy[i])
                                          for i in range(len(cons_laws_sympy))]

                    replacements.append([dep_conc_in_laws[i][0], '(' + str(temp[0]) + ')'])

            dep_conc_in_laws = self.__dependent_species_concentrations(cons_laws_sympy, dep_spec_conc)

            if self.__is_list_empty(dep_conc_in_laws):
                flag = False

        return replacements    

    def __building_ode_str(self, replacements, ind_spec_conc, indp_odes):
        indp_odes_str = []
        # making the replacements in the indep. ODEs
        for i in range(len(indp_odes)):
            for j in range(len(replacements)):
                indp_odes[i] = indp_odes[i].subs(replacements[j][0], replacements[j][1])
            indp_odes_str.append(str(indp_odes[i]))

        # replacing all powers with ^ instead of **
        for i in range(len(indp_odes_str)):
            indp_odes_str[i] = indp_odes_str[i].replace('**', '^')

        # building the string of ODEs in Antimony syntax
        ode_str = ''
        for i in range(len(ind_spec_conc)):
            ode_str += 'J' + str(i) + ': -> ' + str(ind_spec_conc[i]) + '; ' + indp_odes_str[i] + ';'

        return ode_str

    def __building_ant_str(self, ode_str, kinetic_con, lhs_cons_laws, var_vals):
        vars_to_initialize = kinetic_con+lhs_cons_laws + [str(self.__concentration_pars[i]) for i in range(self.__N)]

        ant_str = ode_str

        for i in range(len(vars_to_initialize)):
            ant_str += str(vars_to_initialize[i]) + ' = ' + str(var_vals[i]) + ';'

        return ant_str

    def __initialize_variables_in_antimony_string(self, cons_laws_sympy, ode_str, cons_laws_lamb, concentration_vals,
                                                  kinetic_vals, kinetic_con):

        # string representation of variables on lhs of mass cons laws
        lhs_cons_laws = ['C'+str(i+1) for i in range(len(cons_laws_sympy))]

        conservation_law_vals = [cons_laws_lamb[i](*tuple(concentration_vals)) for i in range(len(cons_laws_lamb))]

        var_vals = kinetic_vals + conservation_law_vals + concentration_vals

        # The full Antimony string of system of ODEs
        ant_str = self.__building_ant_str(ode_str, kinetic_con, lhs_cons_laws, var_vals)

        return ant_str




import os
import numpy
import sympy
import sympy.utilities.lambdify
import sys
import numpy.linalg
from .bistability_finder import BistabilityFinder


class SemiDiffusiveApproach:
    """
    Class for constructing variables and methods needed for the semi-diffusive approach.
    """

    def __init__(self, cgraph):
        """
        Initialization of the SemiDiffusiveApproach class.

        See also
        ---------
        crnt4sbml.CRNT.get_semi_diffusive_approach()
        """
        self.__cgraph = cgraph
        self.__g = self.__cgraph.get_graph()
        self.__important_info = ""

        # getting edges and nodes in the order they were added
        self.__g_nodes = self.__cgraph.get_g_nodes()
        self.__g_edges = self.__cgraph.get_g_edges()

        # vars used frequently
        self.__numpy_dtype = None
        self.__N = len(self.__cgraph.get_species())
        self.__R = len(self.__cgraph.get_reactions())
        self.__species = self.__cgraph.get_species()
        self.__reactions = self.__cgraph.get_reactions()
        self.__delta = self.__cgraph.get_deficiency()
        self.__M = len(self.__cgraph.get_complexes())

        self.__var_nothing_index = [self.__g.nodes[n]['label'] for n in self.__g_nodes if
                                    self.__g.nodes[n]['species_bc']]

        if not self.__var_nothing_index:
            raise Exception("A boundary species is not present, injectivity approach cannot be ran!")

        if not self.__cgraph.get_dim_equilibrium_manifold() == 0:
            print("Conservation laws present.")
            print("The injectivity approach cannot be ran!")
            sys.exit()

        # compute necessary matrices
        self.__create_y_r_matrix()
        self.__create_s_to_matrix()

        self.__find_key_species_indices()
        self.__find_non_key_species_indices()
        self.__create_symbolic_jacobian_matrix()
        self.__create_lambda_jacobian_matrix()
        self.__create_symbolic_objective_function()
        self.__create_lambda_objective_function()
        self.__create_symbolic_polynomial_function()
        self.__create_lambda_polynomial_function()
        self.__create_concentration_pars()
        self.__create_decision_vector()
        self.__create_lambda_equality_poly_fun()

    def run_optimization(self, bounds=None, iterations=10, sys_min_val=numpy.finfo(float).eps, seed=0, print_flag=False,
                         numpy_dtype=numpy.float64):
        """
        Function for running the optimization problem for the semi-diffusive approach. Note that there are no bounds
        enforced on species' concentrations as they are automatically restricted to be greater than zero by the theory.

        Parameters
        -----------
            bounds: list of tuples
                A list defining the lower and upper bounds for each variable in the decision vector. Here the reactions
                are allowed to be set to a single value.
            iterations: int
                The number of iterations to run the feasible point method.
            sys_min_val: float
                The value that should be considered zero for the optimization problem.
            seed: int
                Seed for the random number generator. None should be used if a random generation is desired.
            print_flag: bool
                Should be set to True if the user wants the objective function values found in the optimization problem
                and False otherwise.
            numpy_dtype:
                The numpy data type used within the optimization routine. All variables in the optimization routine will
                be converted to this data type.
        Returns
        --------
        params_for_global_min: list of numpy arrays
            A list of numpy arrays that correspond to the decision vectors of the problem.
        obj_fun_val_for_params: list of floats
            A list of objective function values produced by the corresponding decision vectors in params_for_global_min.

        Examples
        ---------
        See :ref:`quickstart-injectivity-label` and :ref:`my-injectivity-label`.
        """
        temp_p = numpy.zeros(self.__N, numpy_dtype)
        self.__numpy_dtype = numpy_dtype

        # testing to see if there are any equalities in bounds
        equality_bounds_indices = []
        for i in range(len(bounds)):
            if not isinstance(bounds[i], tuple):
                equality_bounds_indices.append(i)
        
        # recasting user provided input to numpy_dtype
        for i in range(len(bounds)):
            bounds[i] = self.__numpy_dtype(bounds[i])
        sys_min_val = self.__numpy_dtype(sys_min_val)

        if equality_bounds_indices:
            print("Equalities in bounds is not allowed for injectivity approach!")
            sys.exit()

        params_for_global_min, obj_fun_val_for_params, self.__important_info = BistabilityFinder.run_optimization(
            bounds, iterations, sys_min_val, temp_p, self.__penalty_objective_func, self.__feasible_point_check,
            self.__objective_function_to_optimize, self.__final_constraint_check, seed, equality_bounds_indices,
            print_flag, numpy_dtype, [])

        return params_for_global_min, obj_fun_val_for_params

    def run_continuity_analysis(self, species=None, parameters=None, dir_path="./num_cont_graphs",
                                print_lbls_flag=False, auto_parameters=None):
        """
        Function for running the numerical continuation and bistability analysis portions of the semi-diffusive
        approach.

        Parameters
        ------------
            species: string
                A string stating the species that is the y-axis of the bifurcation diagram.
            parameters: list of numpy arrays
                A list of numpy arrays corresponding to the decision vectors that produce a small objective function
                value.
            dir_path: string
                A string stating the path where the bifurcation diagrams should be saved.
            print_lbls_flag: bool
                If True the routine will print the special points found by AUTO 2000 and False will not print any
                special points.
            auto_parameters: dict
                Dictionary defining the parameters for the AUTO 2000 run. Please note that one should **not** set
                'SBML' or 'ScanDirection' in these parameters as these are automatically assigned. It is absolutely
                necessary to set PrincipalContinuationParameter in this dictionary. For more information on these
                parameters refer to :download:`AUTO parameters <../auto2000_input.pdf>`. 'NMX' will default to
                10000 and 'ITMX' to 100.
        Returns
        ---------
            multistable_param_ind: list of integers
                A list of those indices in 'parameters' that produce multistable plots.

        Example
        ---------
        See :ref:`quickstart-injectivity-label` and :ref:`my-injectivity-label`.
        """
        # setting defaults for AUTO
        if 'NMX' not in auto_parameters.keys():
            auto_parameters['NMX'] = 10000
            
        if 'ITMX' not in auto_parameters.keys():
            auto_parameters['ITMX'] = 100

        # making the directory if it doesn't exist
        if not os.path.isdir(dir_path):
            os.mkdir(dir_path)

        species_num = self.__species.index(species) + 1
        species_y = str(self.__species[species_num-1])

        multistable_param_ind, important_info = BistabilityFinder.run_continuity_analysis(species_num, parameters,
                                                                                          self.__initialize_ant_string,
                                                                                          self.__finalize_ant_string,
                                                                                          species_y, dir_path,
                                                                                          print_lbls_flag,
                                                                                          auto_parameters)

        self.__important_info += important_info

        return multistable_param_ind

    def run_greedy_continuity_analysis(self, species=None, parameters=None, dir_path="./num_cont_graphs",
                                       print_lbls_flag=False, auto_parameters=None):
        """
        Function for running the greedy numerical continuation and bistability analysis portions of the semi-diffusive
        approach. This routine uses the initial value of the principal continuation parameter to construct AUTO
        parameters and then tests varying fixed step sizes for the continuation problem. Note that this routine may
        produce jagged or missing sections in the plots provided. To produce better plots one should use the information
        provided by this routine to run :func:`crnt4sbml.SemiDiffusiveApproach.run_continuity_analysis`.

        Parameters
        ------------
            species: string
                A string stating the species that is the y-axis of the bifurcation diagram.
            parameters: list of numpy arrays
                A list of numpy arrays corresponding to the decision vectors that produce a small objective function
                value.
            dir_path: string
                A string stating the path where the bifurcation diagrams should be saved.
            print_lbls_flag: bool
                If True the routine will print the special points found by AUTO 2000 and False will not print any
                special points.
            auto_parameters: dict
                Dictionary defining the parameters for the AUTO 2000 run. Please note that only the
                PrincipalContinuationParameter in this dictionary should be defined, no other AUTO parameters should
                be set. For more information on these parameters refer to :download:`AUTO parameters <../auto2000_input.pdf>`.
        Returns
        ---------
            multistable_param_ind: list of integers
                A list of those indices in 'parameters' that produce multistable plots.

        Example
        ---------
        See :ref:`my-injectivity-label`.
        """
        # setting defaults for AUTO
        if 'NMX' not in auto_parameters.keys():
            auto_parameters['NMX'] = 10000

        if 'ITMX' not in auto_parameters.keys():
            auto_parameters['ITMX'] = 100

        # making the directory if it doesn't exist
        if not os.path.isdir(dir_path):
            os.mkdir(dir_path)

        species_num = self.__species.index(species) + 1
        species_y = str(self.__species[species_num - 1])

        multistable_param_ind, important_info = BistabilityFinder.run_greedy_continuity_analysis\
            (species_num, parameters, self.__initialize_ant_string, self.__finalize_ant_string, species_y, dir_path,
             print_lbls_flag, auto_parameters)

        self.__important_info += important_info

        return multistable_param_ind

    def __initialize_ant_string(self, species_num, pcp_x_reaction):
        self.__create_reaction_rates_vector()

        self.__inflow_vector = sympy.zeros(self.__N, 1)

        edges_inflow = [list(self.__g_edges).index(e) for e in self.__g_edges if e[0] in self.__var_nothing_index]

        [self.__p_symbols.append(sympy.Symbol('p' + str(i))) for i in edges_inflow]

        count = len(self.__edges_true_outflow)
        for i in self.__key_species_indices:
            self.__inflow_vector[i] = self.__p_symbols[count]
            count += 1

        self.__ode = self.__inflow_vector + self.__S_to*self.__reaction_rates_vector

        ode_str = 'var species ' + str(self.__species[species_num-1])
        for i in range(self.__N):
            if self.__species[i] != self.__species[species_num-1]:
                ode_str += ',' + str(self.__species[i])
        ode_str += '; ' 

        for i in range(self.__N):
            ode_str += 'J' + str(i) + ': -> ' + str(self.__species[i]) + '; ' + str(self.__ode[i]) + '; ' 

        # replacing any powers with ^ instead of **
        ode_str = ode_str.replace('**', '^')

        self.__S_to_numpy = numpy.array(self.__S_to).astype(numpy.float64)
        
        # finding the PCP in terms of P_symbols rather than reaction
        edges_indices = self.__edges_true_outflow + edges_inflow
        edge_labels = [self.__g.edges[list(self.__g_edges)[i]]['label'] for i in edges_indices]
        pcp_index = edge_labels.index(pcp_x_reaction)
        pcp_x = str(self.__p_symbols[pcp_index])

        return ode_str, pcp_x

    def __create_reaction_rates_vector(self):
        source_species = [self.__g.nodes[list(self.__g_nodes)[i]]['species'] for i in self.__sources_true_outflow]
        
        source_stoichiometries = [self.__g.nodes[list(self.__g_nodes)[i]]['stoichiometries']
                                  for i in self.__sources_true_outflow]

        self.__p_symbols = [sympy.Symbol('p' + str(i)) for i in self.__edges_true_outflow]

        self.__reaction_rates_vector = sympy.zeros(len(self.__p_symbols), 1)
        for i in range(len(self.__p_symbols)):
            temp_list = [sympy.Symbol(source_species[i][j])**source_stoichiometries[i][j]
                         for j in range(len(source_species[i]))]
            temp_val = 1
            for j in range(len(source_species[i])):
                temp_val *= temp_list[j]
            self.__reaction_rates_vector[i] = self.__p_symbols[i]*temp_val

    def __finalize_ant_string(self, x, ode_str):
        p = -self.__S_to_numpy.dot(x)

        for i in range(len(x)):
            ode_str += str(self.__p_symbols[i]) + ' = ' + str(x[i]) + ';'
        temp = [p[i] for i in self.__key_species_indices]
        count = 0
        
        for i in range(len(x), len(self.__p_symbols)):
            ode_str += str(self.__p_symbols[i]) + ' = ' + str(temp[count]) + ';'
            count += 1 

        lam_ode = []
        temp_sym = self.__p_symbols[:]
        temp_sym += self.__concentration_pars[:]

        for i in range(len(self.__ode)):
            lam_ode.append(sympy.utilities.lambdify(temp_sym, self.__ode[i], 'numpy'))

        for i in range(self.__N):
            ode_str += str(self.__species[i]) + ' = ' + str(1.0) + ';'

        return ode_str

    def __penalty_objective_func(self, x_initial, temp_p, penalty_bounds, equality_bounds_indices, x,
                                 non_equality_bounds_indices, concentration_bounds):
        equality_constraints = numpy.zeros(len(self.__mus_solved_for), dtype=self.__numpy_dtype)
        for i in range(len(self.__mus_solved_for)):
            equality_constraints[i] = self.__lambda_equality_poly_fun[i](*tuple(x_initial))
            
        x = numpy.zeros(len(self.__mu_vec), dtype=self.__numpy_dtype)
        count = 0 
        for j in self.__mus_solved_for_indices:
            x[j] = equality_constraints[count]
            count += 1 

        count = 0 
        for j in self.__decision_vector_indices:
            x[j] = x_initial[count]
            count += 1 

        temp = numpy.zeros(len(self.__key_species_indices), dtype=self.__numpy_dtype)
        count = 0
        for i in self.__key_species_indices:
            temp[count] = numpy.maximum(self.__numpy_dtype(0.0), -self.__lambda_poly_func[i](*tuple(x)))**2
            count += 1

        sum1 = numpy.sum(temp)

        temp = numpy.zeros(len(self.__non_key_species_indices), dtype=self.__numpy_dtype)
        count = 0
        for j in self.__mus_solved_for_indices:
            temp[count] = numpy.maximum(self.__numpy_dtype(0.0), -x[j])**2
            count += 1

        sum2 = numpy.sum(temp)

        sum0 = self.__numpy_dtype(0.0)
        for j in non_equality_bounds_indices:

            sum0 += numpy.maximum(self.__numpy_dtype(0.0), penalty_bounds[j][0]-x_initial[j])**2

            sum0 += numpy.maximum(self.__numpy_dtype(0.0), x_initial[j] - penalty_bounds[j][1])**2

        sumval = sum1+sum2 + sum0

        return sumval 

    def __feasible_point_check(self, x_initial, result_fun, sys_min_val, equality_bounds_indices,
                               non_equality_bounds_indices, penalty_bounds, concentration_bounds):
        finite_chk = numpy.isfinite(x_initial)
        if numpy.all(finite_chk):
            
            test = []
            for j in non_equality_bounds_indices:
                test.append(x_initial[j] >= penalty_bounds[j][0] and x_initial[j] <= penalty_bounds[j][1])
            boundry_chk = numpy.all(test)

            equality_constraints = numpy.zeros(len(self.__mus_solved_for), dtype=self.__numpy_dtype)
            all_constraints = []
            for i in range(len(self.__mus_solved_for)):
                equality_constraints[i] = self.__lambda_equality_poly_fun[i](*tuple(x_initial))
                all_constraints.append(equality_constraints[i])

            x = numpy.zeros(len(self.__mu_vec), dtype=self.__numpy_dtype)
            count = 0
            for j in self.__mus_solved_for_indices:
                x[j] = equality_constraints[count]
                count += 1

            count = 0
            for j in self.__decision_vector_indices:
                x[j] = x_initial[count]
                count += 1    

            for i in self.__key_species_indices:  
                all_constraints.append(self.__lambda_poly_func[i](*tuple(x)))

            constraints_chk = all([i > self.__numpy_dtype(0.0) for i in all_constraints])

            # putting the feasible points in x_candidates
            if (abs(result_fun) <= self.__numpy_dtype(1e-200)) and boundry_chk and constraints_chk: 
                return [True, x_initial]
            else:
                return [False, x_initial]
        else:
            return [False, x_initial]

    def __objective_function_to_optimize(self, x_initial, temp_c, penalty_bounds, sys_min_val, equality_bounds_indices,
                                         x, non_equality_bounds_indices, concentration_bounds):
        test = []
        for j in non_equality_bounds_indices:
            test.append(x_initial[j] >= penalty_bounds[j][0] and x_initial[j] <= penalty_bounds[j][1])
        boundry_chk = numpy.all(test)

        if boundry_chk:

            equality_constraints = numpy.zeros(len(self.__mus_solved_for), dtype=self.__numpy_dtype)
            all_constraints = []
            for i in range(len(self.__mus_solved_for)):
                equality_constraints[i] = self.__lambda_equality_poly_fun[i](*tuple(x_initial))
                all_constraints.append(equality_constraints[i])

            x = numpy.zeros(len(self.__mu_vec), dtype=self.__numpy_dtype)
            count = 0
            for j in self.__mus_solved_for_indices:
                x[j] = equality_constraints[count]
                count += 1

            count = 0
            for j in self.__decision_vector_indices:
                x[j] = x_initial[count]
                count += 1

            for i in self.__key_species_indices:
                all_constraints.append(self.__lambda_poly_func[i](*tuple(x)))

            constraints_chk = all([i > self.__numpy_dtype(0.0) for i in all_constraints])

            if constraints_chk:
                return self.__lambda_objective_fun(*tuple(x))
            else:
                return numpy.PINF
        else:
            return numpy.PINF

    def __final_constraint_check(self, x_initial, penalty_bounds, sys_min_val, equality_bounds_indices,
                                 concentration_bounds):
        non_equality_bounds_indices = [i for i in range(len(penalty_bounds)) if i not in equality_bounds_indices]

        test = []
        for j in non_equality_bounds_indices:
            test.append(x_initial[j] >= penalty_bounds[j][0] and x_initial[j] <= penalty_bounds[j][1])
        boundry_chk = numpy.all(test)

        equality_constraints = numpy.zeros(len(self.__mus_solved_for), dtype=self.__numpy_dtype)
        all_constraints = []
        for i in range(len(self.__mus_solved_for)):
            equality_constraints[i] = self.__lambda_equality_poly_fun[i](*tuple(x_initial))
            all_constraints.append(equality_constraints[i])

        x = numpy.zeros(len(self.__mu_vec), dtype=self.__numpy_dtype)
        count = 0
        for j in self.__mus_solved_for_indices:
            x[j] = equality_constraints[count]
            count += 1

        count = 0
        for j in self.__decision_vector_indices:
            x[j] = x_initial[count]
            count += 1

        for i in self.__key_species_indices:
            all_constraints.append(self.__lambda_poly_func[i](*tuple(x)))

        constraints_chk = all([i > self.__numpy_dtype(0.0) for i in all_constraints])

        # must convert x to numpy.float64 because higher
        # is not supported in linalg
        x_converted = numpy.float64(x)

        rank_jacobian = numpy.linalg.matrix_rank(self.__lambda_J(*tuple(x_converted)))

        rank_jacobian_chk = rank_jacobian == self.__N - 1

        if constraints_chk and boundry_chk and rank_jacobian_chk:
            return [True, x]

        else:
            return [False, []]

    def __create_y_r_matrix(self):
        self.__sources_true_outflow = [list(self.__g_nodes).index(e[0]) for e in self.__g_edges if e[0]
                                       not in self.__var_nothing_index]
        self.__Y_r = sympy.zeros(self.__N, len(self.__sources_true_outflow))

        for i in range(len(self.__sources_true_outflow)):
            self.__Y_r[:, i] = self.__cgraph.get_y()[:, self.__sources_true_outflow[i]]

    def __create_s_to_matrix(self):
        self.__edges_true_outflow = [list(self.__g_edges).index(e) for e in self.__g_edges if
                                     (e[1] in self.__var_nothing_index) or (e[0] not in self.__var_nothing_index)]

        self.__S_to = sympy.zeros(self.__N, len(self.__edges_true_outflow))

        for i in range(len(self.__edges_true_outflow)): 
            self.__S_to[:, i] = self.__cgraph.get_s()[:, self.__edges_true_outflow[i]]

    def __create_symbolic_jacobian_matrix(self):
        self.__mu_vec = [sympy.symbols('v_' + str(i+1)) for i in self.__edges_true_outflow]

        diag_mu = sympy.diag(*self.__mu_vec)

        self.__symbolic_J = self.__S_to*diag_mu*self.__Y_r.T

    def __create_lambda_jacobian_matrix(self):
        self.__lambda_J = sympy.utilities.lambdify(self.__mu_vec, self.__symbolic_J, 'numpy')

    def __create_symbolic_objective_function(self):
        self.__symbolic_objective_fun = (self.__symbolic_J.det(method='lu'))**2

    def __create_lambda_objective_function(self):
        self.__lambda_objective_fun = sympy.utilities.lambdify(self.__mu_vec, self.__symbolic_objective_fun, 'numpy')

    def __create_symbolic_polynomial_function(self):
        mu_sym_mat = sympy.Matrix([[mu] for mu in self.__mu_vec])
        self.__symbolic_poly_func = -self.__S_to*mu_sym_mat
        
    def __create_lambda_polynomial_function(self):
        self.__lambda_poly_func = []
        for i in range(self.__N):
            self.__lambda_poly_func += [sympy.utilities.lambdify(self.__mu_vec, self.__symbolic_poly_func[i], 'numpy')]

    def __create_lambda_equality_poly_fun(self):
        self.__lambda_equality_poly_fun = []
        for i in range(len(self.__mus_solved_for)):
            self.__lambda_equality_poly_fun += [sympy.utilities.lambdify(self.__decision_vector,
                                                                         self.__symbolic_equality_poly_fun[i][0],
                                                                         'numpy')]

    def __find_key_species_indices(self):
        key_species_vertex = []
        for e in self.__g_edges:
            if e[0] in self.__var_nothing_index:
                key_species_vertex.append(e[1])

        # putting key_species in order
        key_species = [i for i in self.__species if i in key_species_vertex]

        self.__key_species_indices = [self.__species.index(i) for i in key_species]

    def __find_non_key_species_indices(self):
        self.__non_key_species_indices = [i for i in range(len(self.__species)) if i not in self.__key_species_indices]

    def __create_concentration_pars(self):
        self.__concentration_pars = [sympy.Symbol(self.__species[i]) for i in range(self.__N)]

    def __create_decision_vector(self):
        atoms_of_p_non_key = []
        for i in self.__non_key_species_indices:
            atoms_of_p_non_key.append(self.__symbolic_poly_func[i].atoms(sympy.Symbol))

        available_mus = list(set.union(*atoms_of_p_non_key))

        # putting them in order for the sake of reproducibility
        available_mus = [i for i in self.__mu_vec if i in available_mus]

        # putting them in order for the sake of reproducibility
        for i in range(len(atoms_of_p_non_key)):
            atoms_of_p_non_key[i] = [j for j in self.__mu_vec if j in atoms_of_p_non_key[i]]

        self.__mus_solved_for = []
        for i in range(len(atoms_of_p_non_key)):
            for j in atoms_of_p_non_key[i]:
                if j in available_mus:
                    if not any([j in atoms_of_p_non_key[ii] for ii in range(len(atoms_of_p_non_key)) if i != ii]):
                    
                        self.__mus_solved_for.append(j)
                        available_mus.remove(j)
                        break
                    
        self.__symbolic_equality_poly_fun = []
        count = 0
        for i in self.__non_key_species_indices:
            self.__symbolic_equality_poly_fun.append(sympy.solve(self.__symbolic_poly_func[i],
                                                                 self.__mus_solved_for[count]))
            count += 1

        variables_in_decision_vector = list(set(self.__mu_vec) - set(self.__mus_solved_for))

        self.__decision_vector = [i for i in self.__mu_vec if i in variables_in_decision_vector]

        true_outflow_reaction_labels = [self.__g.edges[list(self.__g_edges)[i]]['label']
                                        for i in self.__edges_true_outflow]

        self.__decision_vector_indices = [self.__mu_vec.index(i) for i in self.__decision_vector]

        self.__mus_solved_for_indices = [self.__mu_vec.index(i) for i in self.__mus_solved_for]

        self.__decision_vector_reaction_labels = [true_outflow_reaction_labels[i]
                                                  for i in self.__decision_vector_indices]

    # getters
    def get_y_r_matrix(self):
        """
        Returns SymPy matrix representing the :math:`Y_r` matrix. The columns of which correspond to the true and
        outflow reactions of the molecularity matrix.
        :download:`Fig1Cii.xml <../../sbml_files/Fig1Cii.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Cii.xml")
        >>> approach = network.get_semi_diffusive_approach()
        >>> sympy.pprint(approach.get_y_r_matrix())
            ⎡1  0  0  0  1  0  0  0  0  1  0  0  0  0  0  0⎤
            ⎢                                              ⎥
            ⎢1  0  0  0  0  0  0  0  0  0  1  0  0  0  0  0⎥
            ⎢                                              ⎥
            ⎢0  1  0  0  0  0  1  0  0  0  0  0  0  1  0  0⎥
            ⎢                                              ⎥
            ⎢0  0  1  0  1  0  0  0  0  0  0  0  1  0  0  0⎥
            ⎢                                              ⎥
            ⎢0  0  1  0  0  0  0  0  0  0  0  1  0  0  0  0⎥
            ⎢                                              ⎥
            ⎢0  0  0  1  0  0  0  1  0  0  0  0  0  0  1  0⎥
            ⎢                                              ⎥
            ⎣0  0  0  0  0  1  0  0  1  0  0  0  0  0  0  1⎦
        """
        return self.__Y_r

    def get_s_to_matrix(self):
        """
        Returns SymPy matrix representing the :math:`S_{to}` matrix. The columns of which correspond to the true and
        outflow reactions of the stoichiometric matrix.
        :download:`Fig1Cii.xml <../../sbml_files/Fig1Cii.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Cii.xml")
        >>> approach = network.get_semi_diffusive_approach()
        >>> sympy.pprint(approach.get_s_to_matrix())
            ⎡-1  1   0   0   -1  1   0   1   0   -1  0   0   0   0   0   0 ⎤
            ⎢                                                              ⎥
            ⎢-1  1   0   0   0   0   1   0   0   0   -1  0   0   0   0   0 ⎥
            ⎢                                                              ⎥
            ⎢1   -1  0   0   0   0   -1  0   0   0   0   0   0   -1  0   0 ⎥
            ⎢                                                              ⎥
            ⎢0   0   -1  1   -1  1   1   0   2   0   0   0   -1  0   0   0 ⎥
            ⎢                                                              ⎥
            ⎢0   0   -1  1   0   0   0   1   0   0   0   -1  0   0   0   0 ⎥
            ⎢                                                              ⎥
            ⎢0   0   1   -1  0   0   0   -1  0   0   0   0   0   0   -1  0 ⎥
            ⎢                                                              ⎥
            ⎣0   0   0   0   1   -1  0   0   -1  0   0   0   0   0   0   -1⎦
        """
        return self.__S_to    

    def get_symbolic_objective_fun(self):
        """
        Returns SymPy expression for the objective function of the optimization problem. This is the determinant of
        :math:`S_{to}diag(\mu)Y_r^T` squared.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/sbml_file.xml")
        >>> approach = network.get_semi_diffusive_approach()
        >>> approach.get_symbolic_objective_fun()
        """
        return self.__symbolic_objective_fun

    def get_lambda_objective_fun(self):
        """
        Returns a lambda function representation of the objective function of the optimization problem. Here the
        arguments of the lambda function are given by the values provided by
        :func:`crnt4sbml.SemiDiffusiveApproach.get_mu_vector`.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/sbml_file.xml")
        >>> approach = network.get_semi_diffusive_approach()
        >>> approach.get_lambda_objective_fun()
        """
        return self.__lambda_objective_fun

    def get_symbolic_polynomial_fun(self):
        """
        Returns SymPy matrix representing the vector of polynomial functions, :math:`-S_{to} \mu`.
        :download:`Fig1Cii.xml <../../sbml_files/Fig1Cii.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Cii.xml")
        >>> approach = network.get_semi_diffusive_approach()
        >>> sympy.pprint(approach.get_symbolic_polynomial_fun())
            ⎡   v₁ + v₁₁ - v₂ + v₅ - v₆ - v₈    ⎤
            ⎢                                   ⎥
            ⎢        v₁ + v₁₃ - v₂ - v₇         ⎥
            ⎢                                   ⎥
            ⎢        -v₁ + v₁₇ + v₂ + v₇        ⎥
            ⎢                                   ⎥
            ⎢v₁₆ + v₃ - v₄ + v₅ - v₆ - v₇ - 2⋅v₉⎥
            ⎢                                   ⎥
            ⎢        v₁₅ + v₃ - v₄ - v₈         ⎥
            ⎢                                   ⎥
            ⎢        v₁₈ - v₃ + v₄ + v₈         ⎥
            ⎢                                   ⎥
            ⎣        v₁₉ - v₅ + v₆ + v₉         ⎦
        """
        return self.__symbolic_poly_func

    def get_lambda_polynomial_fun(self):
        """
        Returns a list of lambda functions for the vector of polynomial functions. The index of the list corresponds to
        the row in the vector of polynomial functions. Here the arguments of the lambda function are given by the values
        provided by :func:`crnt4sbml.SemiDiffusiveApproach.get_mu_vector`.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/sbml_file.xml")
        >>> approach = network.get_semi_diffusive_approach()
        >>> approach.get_lambda_polynomial_fun()
        """
        return self.__lambda_poly_func

    def get_key_species(self):
        """
        Returns a list of string variables corresponding to the key species.
        :download:`Fig1Cii.xml <../../sbml_files/Fig1Cii.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Cii.xml")
        >>> approach = network.get_semi_diffusive_approach()
        >>> print(approach.get_key_species())
            ['s1', 's2', 's7']
        """
        return [self.__species[i] for i in self.__key_species_indices]

    def get_non_key_species(self):
        """
        Returns a list of string variables corresponding to those species that are not key species.
        :download:`Fig1Cii.xml <../../sbml_files/Fig1Cii.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Cii.xml")
        >>> approach = network.get_semi_diffusive_approach()
        >>> print(approach.get_non_key_species())
            ['s3', 's6', 's8', 's11']
        """
        return [self.__species[i] for i in self.__non_key_species_indices]

    def get_boundary_species(self):
        """
        Returns a list of string variables corresponding to those species that are defined as boundary species.
        :download:`Fig1Cii.xml <../../sbml_files/Fig1Cii.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Cii.xml")
        >>> approach = network.get_semi_diffusive_approach()
        >>> print(approach.get_boundary_species())
            ['s21']
        """
        return [i for i in self.__var_nothing_index]

    def get_decision_vector(self):
        """
        Returns a list of SymPy variables corresponding to the decision vector for the optimization problem.
        :download:`Fig1Cii.xml <../../sbml_files/Fig1Cii.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Cii.xml")
        >>> approach = network.get_semi_diffusive_approach()
        >>> print(approach.get_decision_vector())
            [v_2, v_3, v_4, v_5, v_6, v_7, v_9, v_11, v_13, v_15, v_17, v_18]

        See also
        ----------
        crnt4sbml.SemiDiffusiveApproach.print_decision_vector
        """
        return self.__decision_vector

    def get_mu_vector(self):
        """
        Returns a list of SymPy variables corresponding to the vector of fluxes, :math:`\mu` .
        :download:`Fig1Cii.xml <../../sbml_files/Fig1Cii.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Cii.xml")
        >>> approach = network.get_semi_diffusive_approach()
        >>> print(approach.get_mu_vector())
            [v_1, v_2, v_3, v_4, v_5, v_6, v_7, v_8, v_9, v_11, v_13, v_15, v_16, v_17, v_18, v_19]
        """
        return self.__mu_vec

    def print_decision_vector(self):
        """
        Prints an easily readable form of the decision vector. It first prints the decision vector and then the
        corresponding reaction labels.
        :download:`Fig1Cii.xml <../../sbml_files/Fig1Cii.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Cii.xml")
        >>> approach = network.get_semi_diffusive_approach()
        >>> approach.print_decision_vector()
            Decision vector for optimization:
            [v_2, v_3, v_4, v_5, v_6, v_7, v_9, v_11, v_13, v_15, v_17, v_18]
            Reaction labels for decision vector:
            ['re1r', 're3', 're3r', 're6', 're6r', 're2', 're8', 're17r', 're18r', 're19r', 're21', 're22']
        """
        print("Decision vector for optimization: ")
        print(self.__decision_vector)
        print("")
        print("Reaction labels for decision vector: ")
        print(self.__decision_vector_reaction_labels)
        print("")

    def generate_report(self):
        """
        Prints out helpful details constructed by :func:`crnt4sbml.SemiDiffusiveApproach.run_optimization` and
        :func:`crnt4sbml.SemiDiffusiveApproach.run_continuity_analysis`.

        Example
        --------
        See :ref:`quickstart-injectivity-label` and :ref:`my-injectivity-label`.
        """
        print(self.__important_info)



import sys
sys.path.insert(0, "..")

import crnt4sbml

network = crnt4sbml.CRNT("../sbml_files/HDA_3_2_1.xml")

approach = network.get_advanced_deficiency_approach()

approach.run_higher_deficiency_algorithm()

